"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

..

Unittest for module, drain_swamp.lock_toggle

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_lock_toggle.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_lock_toggle.py

"""

import importlib.util
import logging
import logging.config
import shutil
from collections.abc import Sequence
from pathlib import Path
from unittest.mock import patch

import pytest

from drain_swamp.backend_abc import BackendType
from drain_swamp.backend_setuptools import BackendSetupTools  # noqa: F401
from drain_swamp.constants import (
    LOGGING,
    g_app_name,
)
from drain_swamp.exceptions import MissingRequirementsFoldersFiles
from drain_swamp.lock_toggle import (
    InFile,
    InFiles,
    _create_symlinks_relative,
    _maintain_symlink,
    _postprocess_abspath_to_relpath,
    lock_compile,
    refresh_links,
    unlock_compile,
)

testdata_lock_compile = (
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        (
            "requirements/prod.in",
            "requirements/pip.in",
            "requirements/pip-tools.in",
            "requirements/dev.in",
            "requirements/manage.in",
            "docs/requirements.in",
        ),
        (),
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        (
            "requirements/prod.in",
            "requirements/pip.in",
            "requirements/pip-tools.in",
            "requirements/dev.in",
            "requirements/manage.in",
            "docs/requirements.in",
            "ci/tox.in",
        ),
        (Path("ci"),),
    ),
)
ids_lock_compile = (
    "Without extra folders",
    "With additional folder, ci",
)


@pytest.mark.skipif(
    importlib.util.find_spec("piptools") is None,
    reason="uses pip-compile part of pip-tools",
)
@pytest.mark.parametrize(
    "path_config, seq_create_these, additional_folders",
    testdata_lock_compile,
    ids=ids_lock_compile,
)
def test_lock_compile(
    path_config,
    seq_create_these,
    additional_folders,
    tmp_path,
    prepare_folders_files,
    prep_pyproject_toml,
    caplog,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_lock_compile" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    # prepare (required and optionals .in files)
    prepare_folders_files(seq_create_these, tmp_path)
    path_config_dest = prep_pyproject_toml(path_config, tmp_path)

    inst = BackendType.load_factory(
        path_config_dest,
        parent_dir=tmp_path,
        additional_folders=additional_folders,
    )

    expected = list(inst.in_files())
    expected_count = len(expected)

    # dry run. Thank you /bin/true
    with (patch(f"{g_app_name}.lock_toggle.PATH_PIP_COMPILE", Path("/bin/true")),):
        gen_lock_files = lock_compile(inst)
        actual_count = len(list(gen_lock_files))
        assert actual_count == 0

    # lock_compile
    gen_lock_files = lock_compile(inst)
    actual_count = len(list(gen_lock_files))
    # assert has_logging_occurred(caplog)
    assert expected_count == actual_count


testdata_resolve_one_iteration = (
    (
        (
            Path("requirements/prod.in"),
            Path("requirements/pins.in"),
            Path("requirements/pip.in"),
        ),
        0,
    ),
    (
        (
            Path("requirements/manage.in"),
            Path("requirements/tox.in"),
            Path("requirements/prod.in"),
            Path("requirements/pins.in"),
        ),
        1,
    ),
)
ids_resolve_one_iteration = (
    "resolves within one loop",
    "takes two loops. After one loop, manage.in not resolved yet",
)


@pytest.mark.parametrize(
    "relpath_files, expected_unresolved",
    testdata_resolve_one_iteration,
    ids=ids_resolve_one_iteration,
)
def test_resolve_one_iteration(
    relpath_files,
    expected_unresolved,
    path_project_base,
    caplog,
    has_logging_occurred,
):
    """Does one loop iteration. Unresolved constraints are expected"""
    # pytest --showlocals --log-level INFO -k "test_resolve_one_iteration" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    # prepare
    abspath_files = []
    path_cwd = path_project_base()
    for p_f in relpath_files:
        abspath = path_cwd.joinpath(p_f)
        abspath_files.append(abspath)

    # act
    files = InFiles(path_cwd, abspath_files)

    #    no resolution has been attempted yet
    files.move_zeroes()

    unresolved_before_count = len(files._files)

    #    Check this test is non-trivial. There is resolving work to be done
    assert unresolved_before_count != 0

    files.resolve_zeroes()
    unresolved_after_count = len(files._files)

    #    Check resolve resolved everything
    assert unresolved_after_count == expected_unresolved

    # assert has_logging_occurred(caplog)
    pass


testdata_resolve_loop = (
    (
        Path("requirements/prod.in"),
        Path("requirements/pins.in"),
        Path("requirements/pip.in"),
    ),
    (
        Path("requirements/manage.in"),
        Path("requirements/tox.in"),
        Path("requirements/prod.in"),
        Path("requirements/pins.in"),
    ),
    pytest.param(
        (
            Path("requirements/manage.in"),
            Path("requirements/prod.in"),
            Path("requirements/pins.in"),
        ),
        marks=pytest.mark.xfail(raises=MissingRequirementsFoldersFiles),
    ),
)
ids_resolve_loop = (
    "prod 0 pins 0 pip 1",
    "prod 0 pins 0 tox 1 manage 3",
    "tox constraint file missing",
)


@pytest.mark.parametrize(
    "relpath_files",
    testdata_resolve_loop,
    ids=ids_resolve_loop,
)
def test_resolve_loop(
    relpath_files,
    path_project_base,
    caplog,
    has_logging_occurred,
):
    """Resolve constraints raising error is unresolvable constraints"""
    # pytest --showlocals --log-level INFO -k "test_resolve_loop" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    assert isinstance(relpath_files, Sequence)

    # prepare
    abspath_files = []
    path_cwd = path_project_base()
    for p_f in relpath_files:
        abspath = path_cwd.joinpath(p_f)
        abspath_files.append(abspath)

    # act
    files = InFiles(path_cwd, abspath_files)
    files.resolution_loop()
    #  No remaining InFile with unresolved constraints
    assert len(files._files) == 0

    # assert has_logging_occurred(caplog)
    pass


testdata_infile = (
    (
        Path("requirements/pip.in"),
        ("requirements/pins.in",),
        ("pip", "setuptools", "setuptools-scm"),
    ),
)
ids_infile = ("pip.in 1 constraint 3 requirements",)


@pytest.mark.parametrize(
    "relpath, constraints, requirements",
    testdata_infile,
    ids=ids_infile,
)
def test_infile(
    relpath,
    constraints,
    requirements,
    path_project_base,
):
    """InFiles calls check_path before call to InFile, which does no checks"""
    # pytest --showlocals --log-level INFO -k "test_infile" tests
    # prepare
    path_cwd = path_project_base()
    set_constraints = {cons for cons in constraints}
    set_requirements = {reqs for reqs in requirements}

    # successful call
    in_ = InFile(relpath, relpath.stem, set_constraints, set_requirements)

    # act
    path_abs = in_.abspath(path_cwd)
    assert path_abs.is_absolute()
    assert path_abs.relative_to(path_cwd) == Path(in_.relpath)

    # eq -- str
    assert in_ == relpath
    assert in_ == str(relpath)

    # eq -- InFile
    set_ins = set()
    set_ins.add(in_)
    assert in_ == set_ins.pop()

    # eq -- unsupported type
    assert in_ != 1.1234


def test_cls_infiles_exceptions(
    tmp_path,
    path_project_base,
):
    # pytest --showlocals --log-level INFO -k "test_cls_infiles" tests
    path_cwd = path_project_base()
    # Expecting a sequence or sequence contains unsupported type
    invalids_types = (
        None,
        1.2345,
        (None,),
        (1.2345,),
        ("Hello World!",),
    )
    for invalid in invalids_types:
        with pytest.raises(TypeError):
            InFiles(tmp_path, invalid)

    abspath_file_0 = tmp_path.joinpath("does-not-exist.in")
    with pytest.raises(FileNotFoundError):
        in_files = (abspath_file_0,)
        InFiles(tmp_path, in_files)

    # file not relative to package base folder
    abspath_file_0.touch()
    with pytest.raises(ValueError):
        in_files = (abspath_file_0,)
        InFiles(path_cwd, in_files)

    # constraints file does not exist
    abspath_file_0.write_text("-c requirements/secrets-to-time-travel.in\n\n")
    in_files = (abspath_file_0,)
    with pytest.raises(MissingRequirementsFoldersFiles):
        InFiles(tmp_path, in_files)


testdata_methods_infiles = (
    (
        Path("requirements/manage.in"),
        Path("requirements/tox.in"),
        Path("requirements/prod.in"),
        Path("requirements/pins.in"),
    ),
)
ids_methods_infiles = ("resolvable requirements constraints",)


@pytest.mark.parametrize(
    "relpath_files",
    testdata_methods_infiles,
    ids=ids_methods_infiles,
)
def test_methods_infiles(
    relpath_files,
    path_project_base,
):
    # pytest --showlocals --log-level INFO -k "test_methods_infiles" tests
    assert isinstance(relpath_files, Sequence)

    # prepare
    abspath_files = []
    path_cwd = path_project_base()
    for p_f in relpath_files:
        abspath = path_cwd.joinpath(p_f)
        abspath_files.append(abspath)

    files = InFiles(path_cwd, abspath_files)
    check_files = (
        None,
        "files",
    )
    check_fors = (
        "requirements/tox.in",
        Path("requirements/tox.in"),
    )
    for checks_files in check_files:
        for check_for in check_fors:
            # InFiles.in_generic
            assert files.in_generic(check_for, set_name=checks_files) is True
            # InFiles.__contains__
            assert check_for in files
            # InFiles.get_by_relpath
            in_ = files.get_by_relpath(check_for, set_name=checks_files)
            assert isinstance(in_, InFile)

        # unsupported type
        invalids = (
            None,
            ("requirements/goose-meat.in",),
        )
        for invalid in invalids:
            with pytest.raises(ValueError):
                files.get_by_relpath(invalid, set_name=checks_files)

        # nonexistent file
        check_for = "requirements/goose-meat.in"
        in_ = files.get_by_relpath(check_for, set_name=checks_files)
        assert in_ is None


testdata_unlock_compile = (
    pytest.param(
        Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        ),
        (
            "requirements/prod.in",
            "requirements/pins.in",
            "requirements/pip.in",
            "requirements/manage.in",
        ),
        (),
        marks=pytest.mark.xfail(raises=MissingRequirementsFoldersFiles),
    ),
    (
        Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        ),
        (
            "requirements/prod.in",
            "requirements/pins.in",
            "requirements/tox.in",
            "requirements/manage.in",
        ),
        (),
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        (
            "requirements/prod.in",
            "requirements/pins.in",
            "requirements/pip.in",
            "requirements/pip-tools.in",
            "requirements/dev.in",
            "requirements/tox.in",
            "requirements/manage.in",
            "docs/requirements.in",
        ),
        (),
    ),
)
ids_unlock_compile = (
    "missing tox",
    "manage and tox",
    "constraint path needs to be resolved",
)


@pytest.mark.parametrize(
    "path_config, seq_create_these, additional_folders",
    testdata_unlock_compile,
    ids=ids_unlock_compile,
)
def test_unlock_compile(
    path_config,
    seq_create_these,
    additional_folders,
    tmp_path,
    prepare_folders_files,
    prep_pyproject_toml,
    path_project_base,
    caplog,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_unlock_compile" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    # prepare
    assert isinstance(seq_create_these, Sequence)
    #    makes both folders and files, but blank files
    prepare_folders_files(seq_create_these, tmp_path)

    #    move real files, no need to create folders
    path_cwd = path_project_base()
    for p_f in seq_create_these:
        abspath_src = path_cwd.joinpath(p_f)
        abspath_dest = tmp_path.joinpath(p_f)
        shutil.copy2(abspath_src, abspath_dest)
        # abspath_files.append(abspath_dest)
        pass

    path_config_dest = prep_pyproject_toml(path_config, tmp_path)

    # prepare (required and optionals .in files)
    inst = BackendType.load_factory(
        path_config_dest,
        parent_dir=tmp_path,
        additional_folders=additional_folders,
    )
    assert inst.parent_dir == tmp_path

    expected = list(inst.in_files())
    expected_count = len(expected)

    # unlock_compile
    gen = unlock_compile(inst)
    unlock_files = list(gen)
    unlock_files_count = len(unlock_files)
    assert expected_count - 1 == unlock_files_count

    # assert has_logging_occurred(caplog)
    pass


def test_create_symlinks_relative(tmp_path, prepare_folders_files, caplog):
    # pytest --showlocals --log-level INFO -k "test_create_symlinks_relative" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    src_abspath = tmp_path.joinpath("src.py")

    # NotADirectoryError (cwd)
    src = "hi-there"
    dest = "hello.lnk"
    with pytest.raises(NotADirectoryError):
        _create_symlinks_relative(src, dest, src_abspath)

    with pytest.raises(NotADirectoryError):
        _maintain_symlink(src_abspath, src_abspath)

    # FileNotFoundError (src)
    with pytest.raises(FileNotFoundError):
        _create_symlinks_relative(src, dest, tmp_path)

    with pytest.raises(FileNotFoundError):
        _maintain_symlink(tmp_path, src_abspath)

    # ValueError (dest suffix wrong)
    # _maintain_symlink does produce a ValueError
    path_src = tmp_path.joinpath(src)
    path_src.touch()
    dest = "hello.txt"
    with pytest.raises(ValueError):
        _create_symlinks_relative(src, dest, tmp_path)

    # Success case
    src = "requirements/hi-there.unlock"
    src_abspath = tmp_path.joinpath(src)
    dest = "hi-there.lnk"
    prepare_folders_files((src,), tmp_path)
    _create_symlinks_relative(src, dest, tmp_path)
    path_dest_expected = tmp_path.joinpath("requirements", dest)
    is_symlink = path_dest_expected.is_symlink()
    assert is_symlink
    assert path_dest_expected.resolve() == tmp_path.joinpath(src)
    path_dest_expected.unlink()
    assert not path_dest_expected.exists()

    _maintain_symlink(tmp_path, src_abspath)
    is_symlink = path_dest_expected.is_symlink()
    assert is_symlink
    assert path_dest_expected.resolve() == tmp_path.joinpath(src)
    path_dest_expected.unlink()
    assert not path_dest_expected.exists()

    # os.symlink not supported on this platform --> NotImplementedError
    src = "requirements/yo.lock"
    src_abspath = tmp_path.joinpath(src)
    prepare_folders_files((src,), tmp_path)
    dest = "yo.lnk"
    with (
        patch("os.symlink", side_effect=NotImplementedError),
        pytest.raises(NotImplementedError),
    ):
        _create_symlinks_relative(src, dest, tmp_path)
    with (
        patch("os.symlink", side_effect=NotImplementedError),
        pytest.raises(NotImplementedError),
    ):
        _maintain_symlink(tmp_path, src_abspath)


testdata_refresh = (
    pytest.param(
        Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        ),
        (
            "requirements/prod.in",
            "requirements/pins.in",
            "requirements/pip.in",
            "requirements/manage.in",
        ),
        (
            "requirements/prod.lock",
            "requirements/pip.lock",
            "requirements/manage.lock",
            "requirements/prod.unlock",
            "requirements/pip.unlock",
            "requirements/manage.unlock",
        ),
        (
            "requirements/prod.lnk",
            "requirements/pip.lnk",
            "requirements/manage.lnk",
        ),
        marks=pytest.mark.xfail(raises=MissingRequirementsFoldersFiles),
    ),
    pytest.param(
        Path(__file__).parent.joinpath(
            "_bad_files", "static_dependencies.pyproject_toml"
        ),
        (
            "requirements/prod.in",
            "requirements/pins.in",
            "requirements/pip.in",
            "requirements/tox.in",
            "requirements/manage.in",
        ),
        (
            "requirements/prod.lock",
            "requirements/pip.lock",
            "requirements/tox.lock",
            "requirements/manage.lock",
            "requirements/prod.unlock",
            "requirements/pip.unlock",
            "requirements/tox.unlock",
            "requirements/manage.unlock",
        ),
        (
            "requirements/prod.lnk",
            "requirements/pip.lnk",
            "requirements/tox.lnk",
            "requirements/manage.lnk",
        ),
        marks=pytest.mark.xfail(raises=AssertionError),
    ),
    (
        Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        ),
        (
            "requirements/prod.in",
            "requirements/pins.in",
            "requirements/tox.in",
            "requirements/manage.in",
        ),
        (
            "requirements/prod.lock",
            "requirements/tox.lock",
            "requirements/manage.lock",
            "requirements/prod.unlock",
            "requirements/tox.unlock",
            "requirements/manage.unlock",
        ),
        (
            "requirements/prod.lnk",
            "requirements/tox.lnk",
            "requirements/manage.lnk",
        ),
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        (
            "requirements/prod.in",
            "requirements/pins.in",
            "requirements/pip.in",
            "requirements/pip-tools.in",
            "requirements/dev.in",
            "requirements/tox.in",
            "requirements/manage.in",
            "docs/requirements.in",
        ),
        (
            "requirements/prod.lock",
            "requirements/pip.lock",
            "requirements/pip-tools.lock",
            "requirements/dev.lock",
            "requirements/tox.lock",
            "requirements/manage.lock",
            "docs/requirements.lock",
            "requirements/prod.unlock",
            "requirements/pip.unlock",
            "requirements/pip-tools.unlock",
            "requirements/dev.unlock",
            "requirements/tox.unlock",
            "requirements/manage.unlock",
            "docs/requirements.unlock",
        ),
        (
            "requirements/prod.lnk",
            "requirements/pip.lnk",
            "requirements/pip-tools.lnk",
            "requirements/dev.lnk",
            "requirements/tox.lnk",
            "requirements/manage.lnk",
            "docs/requirements.lnk",
        ),
    ),
)
ids_refresh = (
    "missing tox",
    "pyproject.toml dependencies are static, not dynamic",
    "manage and tox",
    "constraint path needs to be resolved",
)


@pytest.mark.parametrize(
    "path_pyproject_toml, seq_create_in_files, seq_create_lock_files, seq_expected",
    testdata_refresh,
    ids=ids_refresh,
)
def test_refresh_links(
    path_pyproject_toml,
    seq_create_in_files,
    seq_create_lock_files,
    seq_expected,
    tmp_path,
    prepare_folders_files,
    prep_pyproject_toml,
    path_project_base,
    caplog,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_refresh_links" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    path_config = prep_pyproject_toml(path_pyproject_toml, tmp_path)

    # prepare
    assert isinstance(seq_create_in_files, Sequence)
    assert isinstance(seq_create_lock_files, Sequence)

    # prepare .in
    prepare_folders_files(seq_create_in_files, tmp_path)
    #    move real files, no need to create folders
    path_cwd = path_project_base()
    for p_f in seq_create_in_files:
        abspath_src = path_cwd.joinpath(p_f)
        abspath_dest = tmp_path.joinpath(p_f)
        shutil.copy2(abspath_src, abspath_dest)
        # abspath_files.append(abspath_dest)
        pass

    # No .lock file
    #    prepare (required and optionals .in files)
    inst = BackendType.load_factory(
        path_config,
        parent_dir=tmp_path,
    )
    assert inst.parent_dir == tmp_path
    with pytest.raises(MissingRequirementsFoldersFiles):
        refresh_links(inst, is_set_lock=True)
    del inst

    # prepare .lock
    prepare_folders_files(seq_create_lock_files, tmp_path)
    #    move real files, no need to create folders
    for p_f in seq_create_lock_files:
        abspath_src = path_cwd.joinpath(p_f)
        abspath_dest = tmp_path.joinpath(p_f)
        shutil.copy2(abspath_src, abspath_dest)
        # abspath_files.append(abspath_dest)
        pass

    # prepare (required and optionals .in files)
    inst = BackendType.load_factory(
        path_config,
        parent_dir=tmp_path,
    )
    assert inst.parent_dir == tmp_path

    invalids = (
        0.1,
        "Hello world",
    )
    for invalid in invalids:
        with pytest.raises(TypeError):
            refresh_links(inst, is_set_lock=invalid)

    valids = (
        None,
        True,
        False,
    )
    for valid in valids:
        refresh_links(inst, is_set_lock=valid)

        # verify symlinks
        for relpath_expected in seq_expected:
            abspath_expected = tmp_path.joinpath(relpath_expected)
            assert abspath_expected.exists() and abspath_expected.is_symlink()
            # clean up symlink
            abspath_expected.unlink()
            assert not abspath_expected.exists()

    # assert has_logging_occurred(caplog)
    pass


def test_postprocess_abspath_to_relpath(tmp_path, prepare_folders_files):
    # pytest --showlocals --log-level INFO -k "test_postprocess_abspath_to_relpath" tests
    # prepare
    #    .in
    seq_create_in_files = (
        "requirements/prod.in",
        "docs/requirements.in",
    )
    prepare_folders_files(seq_create_in_files, tmp_path)

    #    .lock
    lines = (
        "#\n"
        "click==8.1.7\n"
        "    # via\n"
        f"    #   -c {tmp_path!s}/docs/../requirements/prod.in\n"
        "    #   click-log\n"
        "    #   scriv\n"
        "    #   sphinx-external-toc-strict\n"
        "    #   uvicorn\n"
        "sphobjinv==2.3.1.1\n"
        f"    # via -r {tmp_path!s}/docs/requirements.in\n\n"
    )
    path_doc_lock = tmp_path.joinpath("docs", "requirements.lock")
    path_doc_lock.write_text(lines)

    # act
    _postprocess_abspath_to_relpath(path_doc_lock, tmp_path)

    # verify
    #    Within file contents, absolute path of parent folder is absent
    contents = path_doc_lock.read_text()
    is_not_occur_once = str(tmp_path) not in contents
    assert is_not_occur_once is True



"""An integration unittest. **DO NOT** include in :code:`coverage run`.

This unit test must **avoid importing anything from drain_swamp** package.

pytest config file, ``tests/conftest.py``, imports from drain_swamp and
therefore pytest is avoided.

.. warning:: 2nd run

   integration unittest are not run along with 1st run unittests.

   Does not play nice with coverage. Necessary to isolate execution; run seperately.

"""

import subprocess
import sys
import unittest
from pathlib import Path


class TestPep366(unittest.TestCase):
    def setUp(self):
        if "__pycache__" in __file__:
            # cached
            self.path_tests = Path(__file__).parent.parent
        else:
            # not cached
            self.path_tests = Path(__file__).parent
        self.cwd = self.path_tests.parent

    def test_runs_off_source_code(self):
        """Run from source code"""
        cmds = (
            (
                (sys.executable, "src/drain_swamp/cli_igor.py", "list"),
                lambda x: x == 0,
            ),
            (
                (sys.executable, "src/drain_swamp/cli_unlock.py", "is_lock"),
                lambda x: x in [0, 1],
            ),
        )

        for cmd, fcn_exit_compare in cmds:
            try:
                proc = subprocess.run(
                    cmd,
                    shell=False,
                    stdout=subprocess.PIPE,  # click.secho --> sys.stdout
                    stderr=subprocess.PIPE,
                    cwd=self.cwd,  # set cwd. Could have been done within cmd
                    text=True,
                )
            except subprocess.CalledProcessError:  # pragma: no cover
                assert False
            else:
                # str_out = proc.stdout
                # str_err = proc.stderr
                exit_code = proc.returncode
                is_code_match = fcn_exit_compare(exit_code)
                assert is_code_match

    def test_with_relative_path(self):
        """--path [relative path]

        Equivalent to
        :code:`python src/drain_swamp/cli_unlock.py is_lock --path pyproject.toml`
        """
        fcn_exit_compare = lambda x: x in [0, 1]  # noqa: E731

        cmd = (
            sys.executable,
            "src/drain_swamp/cli_unlock.py",
            "is_lock",
            "--path",
            "pyproject.toml",
        )
        try:
            proc = subprocess.run(
                cmd,
                shell=False,  # shell understands the relative path
                stdout=subprocess.PIPE,  # click.secho --> sys.stdout
                stderr=subprocess.PIPE,
                cwd=self.cwd,  # set cwd. Could have been done within cmd
                text=True,
            )
        except subprocess.CalledProcessError:  # pragma: no cover
            assert False
        else:
            # str_out = proc.stdout
            # str_err = proc.stderr
            exit_code = proc.returncode
            is_code_match = fcn_exit_compare(exit_code)
            assert is_code_match


if __name__ == "__main__":  # pragma: no cover
    """See warning about necessity for execution isolation

    .. code-block:: shell

       python -m unittest integration.test_pep366 --locals

       python -m unittest integration.test_pep366 \
       -k TestPep366.test_runs_off_source_code --locals --buffer

       python -m unittest integration.test_pep366 \
       -k TestPep366.test_with_relative_path --locals --buffer

    """
    unittest.main(tb_locals=True)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Utility function to retrieve the global hook_manager singleton.

.. py:data:: _PLUGIN_HOOKS
   :type: str
   :value: "drain_swamp.hooks"

   Entrypoint to load hooks from for built-in or installed plugins

.. py:data:: logger
   :type: logging.Logger

   Module level logger

.. seealso::

   **Avoid** reinventing the wheel

   `SOURCE <https://raw.githubusercontent.com/kedro-org/kedro/main/kedro/framework/hooks/manager.py>`_
   `LICENSE:Apache 2.0 <https://github.com/kedro-org/kedro/blob/main/LICENSE.md>`_

"""

import importlib.util
import inspect
import logging
import pkgutil
import sys
from collections.abc import Sequence
from types import ModuleType
from typing import TYPE_CHECKING

from pluggy import PluginManager

from .constants import (
    _PLUGIN_HOOKS,
    DOTTED_PATH_SPECS,
    HOOK_NAMESPACE,
)

if TYPE_CHECKING:
    from importlib.machinery import ModuleSpec

logger = logging.getLogger(__name__)

__all__ = (
    "before",
    "after",
    "get_plugin_manager",
)


def before(hook_name, hook_impls, kwargs):  # pragma: no cover
    """Hook call tracing. Before a hook is called

    :param hook_name: Name of hook implementation
    :type hook_name: str
    :param hook_impls:

       Sequence of HookImpl instances. Can see which hooks are being executed

    :type hook_impls: collections.abc.Sequence[pluggy.HookImpl]
    :param kwargs: Keyword args passed into hook impl instances
    :type kwargs: collections.abc.Mapping[str, typing.Any]

    Usage

    .. code-block:: text

       from pluggy import PluginManager
       namespace = "ds"
       pm = PluginManager(namespace)
       # add hookspecs
       # register hooks
       undo = pm.add_hookcall_monitoring(before, after)

    .. seealso::

       https://pluggy.readthedocs.io/en/latest/api_reference.html#pluggy.PluginManager.add_hookcall_monitoring


    """
    msg_info = f"trace {hook_name} (before) {hook_impls!r} kwargs {kwargs!r}"
    logger.info(msg_info)


def after(outcome, hook_name, hook_impls, kwargs):  # pragma: no cover
    """Runs after plugin hook execution. Gatherings non-None outcomes into a list

    :param outcome: List of non-None plugin hook results
    :type outcome: pluggy.Result[typing.Any]
    :param hook_name: Name of hook implementation
    :type hook_name: str
    :param hook_impls:

       Sequence of HookImpl instances. Can see which hooks are being executed

    :type hook_impls: collections.abc.Sequence[pluggy.HookImpl]
    :param kwargs: Keyword args passed into hook impl instances
    :type kwargs: collections.abc.Mapping[str, typing.Any]
    """
    msg_info = (
        f"trace {hook_name} (after) {hook_impls!r} kwargs {kwargs!r} "
        f"--> {outcome.get_result()}"
    )
    logger.info(msg_info)


def lazy_dotted_path(dotted_path):
    """Dotted path useful for loading plugin specs

    Should not contain spec classes

    :param dotted_path: Dotted path to module
    :type dotted_path: str
    :returns: Lazy imported module
    :rtype: types.ModuleType
    :raises:

       - :py:exc:`TypeError` -- Expecting valid dotted path

    """
    mod_path = "drain_swamp.monkey.hooks:lazy_dotted_path"
    is_ng = dotted_path is None or not isinstance(dotted_path, str)
    if is_ng:
        msg_warn = (
            f"{mod_path} Expecting valid dotted path to module "
            "containing plugin functions"
        )
        raise TypeError(msg_warn)

    try:
        mod = sys.modules[dotted_path]
    except KeyError:
        spec = importlib.util.find_spec(dotted_path)
        module = importlib.util.module_from_spec(spec)
        loader = importlib.util.LazyLoader(spec.loader)
        # Make module with proper locking and get it inserted into sys.modules.
        loader.exec_module(module)
        mod = module

    return mod


def lazy_package(mod_pkg):
    """Lazy load module

    - fullname is a dotted path

    module should contain only spec functions

    - fullname is a Sequence[ModuleType]

    modules should contain only spec classes module should have a
    ``__all__`` containing only the spec classes names

    :param mod_pkg: Dotted path to package module
    :type mod_pkg: ModuleType
    :returns: Lazy imported module
    :rtype: Sequence[types.ModuleType]
    :raises:

       - :py:exc:`TypeError` -- Unsupported type. Expecting dotted path or ModuleType

    """
    if TYPE_CHECKING:
        mod_path: str
        msg_warn: str
        plugins_dir_paths: Sequence[str]
        mod_specs: list[ModuleSpec]
        module: ModuleType

    mod_path = "drain_swamp.monkey.hooks.manager:lazy_package"
    is_not_module = mod_pkg is None or not inspect.ismodule(mod_pkg)
    if is_not_module:
        msg_warn = f"{mod_path} Expected package module got {type(mod_pkg)}"
        raise TypeError(msg_warn)

    # Should contain only spec classes, no spec functions
    package_plugins = mod_pkg
    """package_plugins package (folder tree) contains builtin plugins' implementations
    location of hook implementations: ..plugins/ds_*
    plugin package ``__init__.py`` is empty, so importing can't raise Exceptions
    onerror param is therefore unneeded
    """
    # Could contain package and sub package paths
    plugins_dir_paths = package_plugins.__path__

    # https://docs.python.org/3/library/importlib.html#importlib.machinery.ModuleSpec
    mod_specs = [
        module_finder.find_spec(name)
        for module_finder, name, _ in pkgutil.walk_packages(plugins_dir_paths)
    ]

    ret = []
    for spec in mod_specs:
        module = importlib.util.module_from_spec(spec)
        loader = importlib.util.LazyLoader(spec.loader)
        # Make module with proper locking and get it inserted into sys.modules.
        loader.exec_module(module)
        mod = module
        ret.append(mod)

    return ret


def _filter_specs(member):  # pragma: no cover
    """Filter function used by inspect. Filter out non-plugins

    :param member: A module member
    :type member: typing.Any
    :returns: True if a module member a class and is a registered
    :rtype: bool
    """
    ret = inspect.isfunction(member) or inspect.isclass(member)
    return ret


def iter_specs(mod, filter_=_filter_specs):  # pragma: no cover
    """Iteratable of pluggy spec or impl classes

    Too hard to inspect spec and impl classes and functions. Using spec
    module ``__all__`` in leiu of inspection or validator.

    :param mod: A package module
    :type mod: types.ModuleType
    :param filter_: A filter function which allows only classes
    :type filter_: collections.abc.Callable[[tying.Any], bool]
    :returns: Plugin classes Iterator
    :rtype: collections.abc.Iterable[typing.Any]
    """
    lst_members = inspect.getmembers(mod, _filter_specs)
    for name, member in lst_members:
        if name in mod.__all__:
            logger.info(f"member (yield): {member!r}")
            yield member
    yield from ()


def _create_hook_manager(namespace, specs_dotted_path):
    """Create a new PluginManager instance and register hook specs.

    :param namespace: hook namespace, so as to support multiple
    :type namespace: str
    :param specs_dotted_path: specs module dotted path
    :type specs_dotted_path: str
    :returns: Plugin manager instance
    :rtype: pluggy.PluginManager
    :meta private:

    .. todo:: lazy import

       Built-in specs are hardcoded. Refactor as lazy imports

    """
    manager = PluginManager(namespace)
    # manager.trace.root.setwriter(logger.debug)
    # undo = manager.enable_tracing()

    # Avoid TypeError
    assert isinstance(specs_dotted_path, str)

    # Load specs given dotted path to spec module
    mod_spec = lazy_dotted_path(specs_dotted_path)

    if isinstance(mod_spec, ModuleType):
        # DOTTED_PATH_SPECS was str, so mod_spec is ModuleType
        manager.add_hookspecs(mod_spec)
    elif isinstance(mod_spec, Sequence):  # pragma: no cover
        # DOTTED_PATH_SPECS was Sequence[ModuleType] --> Sequence[type]
        for kls in iter_specs(mod_spec):
            # types.ModuleType | type
            manager.add_hookspecs(kls)
    else:  # pragma: no cover
        pass

    return manager


def _register_hooks(pm, hooks):
    """Register all hooks as specified in ``hooks`` with the global ``hook_manager``.

    :param hook_manager: Hook manager instance to register the hooks with.
    :type hook_manager: pluggy.PluginManager
    :param hooks: Hooks that need to be registered.
    :type hooks: collections.abc.Iterable[typing.Any | tuple[typing.Any, str | None]]
    :meta private:
    """
    for hooks_collection in hooks:
        # Sometimes users might call hook registration more than once, in which
        # case hooks have already been registered, so we perform a simple check
        # here to avoid an error being raised and break user's workflow.
        if not isinstance(hooks_collection, tuple):
            plugin = hooks_collection
        else:  # pragma: no cover
            # Can pass in a (plugin, plugin name). Not sure what the point is
            plugin = hooks_collection[0]

        # Avoids ValueError
        is_registered = pm.is_registered(plugin)

        if inspect.ismodule(plugin) or inspect.isaclass(plugin):
            # module containing only functions
            # class implementing spec functions as class methods
            if not is_registered:
                pm.register(plugin)
            else:  # pragma: no cover
                pass
        else:  # pragma: no cover
            pass


def _register_hooks_entry_points(hook_manager, entrypoint_plugins, disabled_plugins):
    """Register pluggy hooks from python package entrypoints.

    :param hook_manager:

       Hook manager instance to register the hooks with.

    :type hook_manager: pluggy.PluginManager
    :param entrypoint_plugins:

       Setuptools entrypoint section to specify additional
       :code:`python -m build` time plugins

    :type entrypoint_plugins: str
    :param disabled_plugins:

       An iterable returning the names of plugins which hooks must not
       be registered; any already registered hooks will be unregistered.

    :type disabled_plugins: collections.abc.Iterable[str]
    :meta private:
    """
    already_registered = hook_manager.get_plugins()
    # Method name is misleading:
    # entry points are standard and don't require setuptools,
    # see https://packaging.python.org/en/latest/specifications/entry-points/
    hook_manager.load_setuptools_entrypoints(entrypoint_plugins)
    disabled_plugins = set(disabled_plugins)

    # Get list of plugin/distinfo tuples for all registered plugins.
    plugininfo = hook_manager.list_plugin_distinfo()
    plugin_names = set()
    disabled_plugin_names = set()
    for plugin, dist in plugininfo:
        if dist.project_name in disabled_plugins:
            # `unregister()` is used instead of `set_blocked()` because
            # we want to disable hooks for specific plugin based on project
            # name and not `entry_point` name. Also, we log project names with
            # version for which hooks were registered.
            hook_manager.unregister(plugin=plugin)
            disabled_plugin_names.add(f"{dist.project_name}-{dist.version}")
        elif plugin not in already_registered:
            plugin_names.add(f"{dist.project_name}-{dist.version}")
        else:  # pragma: no cover
            pass

    if len(disabled_plugin_names) != 0:  # pragma: no cover
        str_disabled_plugins = ", ".join(sorted(disabled_plugin_names))
        msg_debug = (f"Hooks are disabled for plugin(s): {str_disabled_plugins}",)
        logger.debug(msg_debug)
    else:  # pragma: no cover
        pass

    if len(plugin_names) != 0:  # pragma: no cover
        str_plugin_names = ", ".join(sorted(plugin_names))
        str_plugin_count = len(plugin_names)
        msg_debug = (
            f"Registered hooks from {str_plugin_count} installed "
            f"plugin(s): {str_plugin_names}"
        )
        logger.debug(msg_debug)
    else:  # pragma: no cover
        pass


def get_plugin_manager(
    mod_pkg_plugins: ModuleType,
    namespace=HOOK_NAMESPACE,
    specs_dotted_path=DOTTED_PATH_SPECS,
    entrypoint_plugins=_PLUGIN_HOOKS,
):
    """Get plugin manager

    - set specs

    - load builtin hooks

    - load entrypoint hooks

    :param namespace:

       Default HOOK_NAMESPACE. hook namespace, so as to support multiple

    :type namespace: str | None
    :param specs_dotted_path: Default DOTTED_PATH_SPECS. dotted path to specs module
    :type specs_dotted_path: str | None
    :param entrypoint_plugins:

       Default _PLUGIN_HOOKS. setuptools entrypoint section to specify
       additional :code:`python -m build` time plugins

    :type entrypoint_plugins: str | None
    :returns: Plugin manager instance
    :rtype: pluggy.PluginManager
    :raises:

      - :py:exc:TypeError -- Expected param mod_pkg_plugins to be a types.ModuleType
        Package folder with modules containg hook implementation functions

    """
    try:
        assert isinstance(mod_pkg_plugins, ModuleType)
    except AssertionError as e:
        msg_exc = (
            "Positional param expected package module. Package folder "
            "with modules containg hook implementation functions"
        )
        raise TypeError(msg_exc) from e

    is_namespace_ng = namespace is None or not isinstance(namespace, str)
    if is_namespace_ng:  # pragma: no cover
        namespace = HOOK_NAMESPACE
    else:  # pragma: no cover
        pass

    is_spec_dotted_path_ng = specs_dotted_path is None or not isinstance(
        specs_dotted_path, str
    )
    if is_spec_dotted_path_ng:  # pragma: no cover
        specs_dotted_path = DOTTED_PATH_SPECS
    else:  # pragma: no cover
        pass

    is_ep_plugins_ng = entrypoint_plugins is None or not isinstance(
        entrypoint_plugins, str
    )
    if is_ep_plugins_ng:  # pragma: no cover
        entrypoint_plugins = _PLUGIN_HOOKS
    else:  # pragma: no cover
        pass

    pm = _create_hook_manager(namespace, specs_dotted_path)

    """package_plugins package (folder tree) contains builtin plugins' implementations
    location of hook implementations: ..plugins/ds_*
    plugin package ``__init__.py`` is empty, so importing can't raise Exceptions
    onerror param is therefore unneeded
    """
    mods = lazy_package(mod_pkg_plugins)
    _register_hooks(pm, mods)

    # dump support for plugin impl klasses
    """
    impls = []
    for mod in mods:
        # plugin impl module
        impls.append(mod)

        for kls in iter_specs(mod):
            log.info(f"{mod_path} plugin name: {kls.__name__!r}")
            # Format for plugins: "{namespace}-{kls.__name__}"
            impls.append((kls(), f"{namespace}-{kls.__name__}"))
        pass
    log.info(f"{mod_path} impls (built-in): {impls!r}")
    _register_hooks(pm, impls)
    """
    pass

    """Occurring during :code:`python -m build`, only way to
    specify more plugins is using entrypoints
    """
    iter_disabled_plugins = ()
    _register_hooks_entry_points(pm, entrypoint_plugins, iter_disabled_plugins)

    return pm



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

package level constants

.. py:data:: __all__
   :type: tuple[str, str, str, str, str, str, str, str, str]
   :value: ("g_app_name", "package_name", "SUFFIX_IN", "SUFFIX_LOCKED", \
   "SUFFIX_UNLOCKED", "SUFFIX_SYMLINK", "PATH_PIP_COMPILE", "PROG_LOCK", \
   "PROG_UNLOCK")

   Module exports

.. py:data:: g_app_name
   :type: str
   :value: "drain_swamp"

   package name. **Not** project name

.. py:data:: package_name
   :type: str
   :value: "drain-swamp"

   g_app_name --> package_name by replacing underscore with hyphen

.. py:data:: SUFFIX_IN
   :type: str
   :value: ".in"

   Uncompiled requirements file suffix

.. py:data:: SUFFIX_LOCKED
   :type: str
   :value: ".lock"

   Allowed locked file, file suffix

.. py:data:: SUFFIX_UNLOCKED
   :type: str
   :value: ".unlock"

   Dependency requirements source files, file suffix. Same as used by pip-lock.

   requirements ``in`` files do not set dependency version restrictions
   unless absolutely unavoidable. In which case, every restriction **must**
   be throughly documented to defend the justification for imposing
   such a restriction so know later whether to keep it or not

.. py:data:: SUFFIX_SYMLINK
   :type: str
   :value: ".lnk"

.. py:data:: PATH_PIP_COMPILE
   :type: pathlib.Path

   Absolute path to pip-compile within venv

.. py:data:: PROG_LOCK
   :type: re.Pattern

   Regex compiled Pattern to find [file name].lock files

.. py:data:: PROG_UNLOCK
   :type: re.Pattern

   Regex compiled Pattern to find [file name].in files

.. py:data:: LOGGING
   :type: dict[str, typing.Any]

   logging dict input to logging.dictConfig

   pytest aware. :code:`"propagate": True,` is required by pytest.
   Otherwise cannot see log messages

   If :code:`DEBUG is True` would require package colorlog. So DEBUG
   is purposefully FALSE

   This logging dict formatter requires package, colorlog. colorized
   debug messages is a feature

   The below code was removed

   .. code-block:: text

      'verbose_with_color': {
          '()': 'colorlog.ColoredFormatter',
          'format': '{log_color}{levelname}{reset} {asctime} {cyan}{name}{reset}: {message}',
          'style': '{',
      },

   .. caution::

      This is example of py38- usage. From py39+, this is no longer valid

      .. code-block:: text

         'root': {
              'handlers': ['console'],
              'level': 'INFO',
         },

   .. seealso::

      `pytest#3697 <https://github.com/pytest-dev/pytest/issues/3697#issuecomment-1984895577>`_

      `pytest#10606 <https://github.com/pytest-dev/pytest/issues/10606#issuecomment-1550123106>`_

"""

import re
import sys
from pathlib import Path

__all__ = (
    "g_app_name",
    "package_name",
    "SUFFIX_IN",
    "SUFFIX_LOCKED",
    "SUFFIX_UNLOCKED",
    "SUFFIX_SYMLINK",
    "PATH_PIP_COMPILE",
    "PROG_LOCK",
    "PROG_UNLOCK",
)

__package__ = "drain_swamp"
g_app_name = "drain_swamp"
# package_name = re.sub(r"\W+", "-", g_app_name.lower())
package_name = g_app_name.lower().replace("_", "-")
SUFFIX_IN = ".in"
SUFFIX_LOCKED = ".lock"
SUFFIX_UNLOCKED = ".unlock"
SUFFIX_SYMLINK = ".lnk"

# Required only for pytest
DEBUG = False
IS_TESTING = "pytest" in sys.modules
formatting = "verbose_with_color" if DEBUG else "simple"
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simple": {
            "format": "{levelname} {name}: {message}",
            "style": "{",
        },
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": formatting,
        },
        "null": {
            "class": "logging.NullHandler",
        },
    },
    "loggers": {
        "django.server": {  # Suppress django HTTP logging because we do it in a middleware
            "handlers": ["console"],
            "level": "WARNING",
            "propagate": IS_TESTING,
        },
        g_app_name: {
            "handlers": ["console"],
            "level": "DEBUG",
            "propagate": IS_TESTING,
        },
    },
}

_PATH_VENV = Path(sys.exec_prefix) / "bin"
PATH_PIP_COMPILE = _PATH_VENV / "pip-compile"

# `regex word+hyphens <https://stackoverflow.com/a/8383339>`_
_pattern_lock = r"\w+(?:-\w+{})+".format(SUFFIX_LOCKED)
_pattern_unlock = r"\w+(?:-\w+{})+".format(SUFFIX_UNLOCKED)
PROG_LOCK = re.compile(_pattern_lock)
PROG_UNLOCK = re.compile(_pattern_unlock)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Get semantic version from scm and write version file

Limit to setuptools-scm codebase. Is exactly equivalent to generic
setuptools-scm ``setup.py``

.. py:data:: log
   :type: logging.Logger

   Module level logger

.. py:data:: __all__
   :type: tuple[str, str, str]
   :value: ("scm_version", "write_to_file", "SEM_VERSION_FALLBACK_SANE")

"""

from __future__ import annotations

import contextlib
import io
import logging
import warnings
from pathlib import Path

from setuptools_scm import (
    Configuration,
    _get_version,
    git,
    hg,
)
from setuptools_scm._integration.pyproject_reading import get_args_for_pyproject
from setuptools_scm._overrides import read_toml_overrides
from setuptools_scm._version_cls import InvalidVersion
from setuptools_scm.fallbacks import parse_pkginfo
from setuptools_scm.version import (
    get_local_node_and_date,
    guess_next_dev_version,
    tag_to_version,
)

from ..version_file.dump_version import write_version_files
from .patch_pyproject_reading import ReadPyproject

log = logging.getLogger("drain_swamp.monkey.wrap_get_version")

__all__ = (
    "scm_version",
    "write_to_file",
    "SEM_VERSION_FALLBACK_SANE",
)

SEM_VERSION_FALLBACK_SANE = "0.0.1"

try_parse = [
    parse_pkginfo,
    git.parse,
    hg.parse,
    git.parse_archival,
    hg.parse_archival,
]


def _parse(root, config):
    """Get the version from underlying scm. Either pkginfo, git, or Mercurial

    :param root: absolute path to root or file on repo base folder
    :type root: str
    :param config:

       A data class usually derived for the most part from ``pyproject.toml``

    :type config: Configuration
    :returns: version str from the underlying scm
    :rtype: ScmVersion | None
    :meta private:
    """
    mod_path = "drain_swamp.monkey.wrap_get_version._parse"
    for maybe_parse in try_parse:
        try:
            # temporarily suppress stderr
            # WARNING  command hg missing:
            # [Errno 2] No such file or directory: 'hg'            _run_cmd.py:195
            with contextlib.redirect_stderr(io.StringIO()):
                parsed = maybe_parse(root, config)
        except OSError as e:
            # Happens often, so suppress importance
            msg_info = f"{mod_path} parse with {maybe_parse!s} failed with: {e!s}"
            log.info(msg_info)
        else:
            if parsed is not None:  # pragma: no cover
                """tox, runs within it's own virtual environment, during
                testing, does not see git. So file finder will not find any match"""
                return parsed
            else:  # pragma: no cover
                pass
    else:
        return None


def scm_version(relative_to, sane_default=SEM_VERSION_FALLBACK_SANE):
    """Get the scm version from git or Mercurial

    Does not write to the version_file

    :param relative_to: absolute path to a file on the repo base folder
    :type relative_to: str
    :param sane_default:

       Default "0.0.1". Fallback version when cannot get scm version
       cuz git uninitialized

    :type sane_default: typing.Any | None
    :returns: parsed version str
    :rtype: str

    .. seealso::

       https://github.com/pypa/setuptools_scm/blob/main/src/setuptools_scm/_get_version_impl.py

    """
    mod_path = "drain_swamp.monkey.wrap_get_version:scm_version"
    is_empty = (
        sane_default is None
        or not isinstance(sane_default, str)
        or len(sane_default.strip()) == 0
    )
    if is_empty:  # pragma: no cover
        sane_default = SEM_VERSION_FALLBACK_SANE
    else:  # pragma: no cover
        pass

    """Configuration.from_file needs to be patched, not Configuration.from_data
    Configuration.from_data does not parse pyproject.toml can't / don't
    write to version_file
    """
    d_data = {
        "parse": _parse,
        "version_scheme": guess_next_dev_version,
        "local_scheme": get_local_node_and_date,
    }
    config = Configuration.from_data(relative_to, d_data)

    scm_ver = _get_version(config, force_write_version_files=False)

    # Apply sane default if no override, no version_file, and no fallback
    if scm_ver is None:
        ret = sane_default
    else:
        ret = scm_ver

    msg_info = f"INFO {mod_path} parsed version: {ret} "
    log.info(msg_info)

    return ret


def write_to_file(
    name: str,
    str_ver: str,
    write_to: str | None = None,
    dist_name: str | None = None,
    is_only_not_exists: bool | None = False,
) -> None:
    """Write version_file

    read_pyproject default is to combine sections:
    [tool.drain_swamp] and [tool.pipenv-unlock]

    Does not read section [tool-setuptools-scm]

    :param name: absolute path to a file on the repo base folder
    :type name: str
    :param str_ver: version str
    :type str_ver: str
    :param write_to: Default None. override version_file. Used for testing
    :type write_to: str | None
    :param dist_name:

       Default None. pyproject.toml section [tool.[name]] Unspecified --> pipenv-unlock

    :type dist_name: str | None
    :param is_only_not_exists: Default False. Write file only if it does not already exit
    :type is_only_not_exists: bool | None
    :raises:

       - :py:exc:`LookupError` -- Either pyproject.toml missing or
         required missing sections

       - :py:exc:`ValueError` -- Semantic version str invalid. Write
         to version file skipped

    """
    if is_only_not_exists is None or not isinstance(is_only_not_exists, bool):
        is_only_not_exists = False
    else:  # pragma: no cover
        pass

    kwargs = {}
    path_file_on_root = Path(name)
    path_root = path_file_on_root.parent

    try:
        # pyproject_data = read_pyproject(path_file_on_root)
        pyproject_data = ReadPyproject()(path=path_file_on_root)
    except LookupError as e:
        msg_info = (
            "To set version_file, in pyproject.toml [tool.pipenv-unlock] "
            "Set version_file to a relative path to a .py file"
        )
        # log.info(msg_info)
        msg_warn = (
            "Either missing pyproject.toml or missing sections: "
            "[tool.drain-swamp] and [tool.pipenv-unlock] "
            f"{path_file_on_root}"
        )
        # log.warning(msg_warn)
        msgs = f"{msg_info}\n{msg_warn}"
        raise LookupError(msgs) from e

    args = get_args_for_pyproject(pyproject_data, dist_name, kwargs)
    args.update(read_toml_overrides(args["dist_name"]))
    # relative_to = args.pop("relative_to", name)

    is_empty = (
        write_to is None or not isinstance(write_to, str) or len(write_to.strip()) == 0
    )
    if is_empty:
        version_file = args.get("version_file")
        str_write_to = None
    else:
        # for testing
        version_file = None
        str_write_to = write_to

    # Check semantic version str. Skip if invalid
    config = Configuration()
    #    py311+ warnings.catch_warnings has category and action kwargs
    with warnings.catch_warnings():
        # setuptools_scm/version.py:102: UserWarning:
        # tag 'golf balls get lost' no version found
        warnings.simplefilter("ignore")
        msg_warn = (
            "drain_swamp.monkey.wrap_get_version invalid semantic version str, "
            f"{str_ver!r}. Write version file skipped"
        )
        # BUG (upstream setuptools-scm): not catching InvalidVersion exception
        try:
            ver_clean = tag_to_version(str_ver, config)
        except InvalidVersion:
            """Passes regex (setuptools_scm.version._parse_version_tag),
            but is invalid semantic version str, e.g. '0.0.1-dev1.g1234123'

            Exception properly handled in setuptools_scm._version_cls._version_as_tuple
            """
            is_invalid = True
        else:
            if ver_clean is None:
                # not a semantic version str e.g.
                # 'getting hit by a golf ball hurts'
                is_invalid = True
            else:
                is_invalid = False

        if is_invalid is True:
            raise ValueError(msg_warn)
        else:
            # str(Version("0.0.1")) --> "0.0.1"
            str_ver_clean = str(ver_clean)
            assert str_ver_clean is not None and isinstance(str_ver_clean, str)
            write_version_files(
                str_ver_clean,
                path_root,
                str_write_to,
                version_file,
                is_only_not_exists=is_only_not_exists,
            )



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

setuptools_scm._integration.pyproject_reading.read_pyproject is too
strict, monkeypatch it!

Without the patch, pyproject.toml [tool.setuptools-scm] section is
missing, raises :py:exc:`LookupError`.

.. warning:: tool_name ordering

   tool_name requires 1st [tool.x] section to be the package name, in
   this case, drain-swamp. e.g. [tool.drain-swamp] section

**CHANGES**

- tool_name (str -> str | Sequence[str])
  from "setuptools-scm"
  to ["drain-swamp", "pipenv-unlock"]

- the tool_name becomes the first element

- combines contents of sections

**USAGE**

.. code-block:: text

   from unittest.mock import patch
   from drain_swamp.monkey.setuptools_scm_pyproject_reading import read_pyproject
   with patch("setuptools_scm._config._read_pyproject", wraps=read_pyproject):
       ...

.. seealso::

   https://github.com/pypa/setuptools_scm/blob/main/src/setuptools_scm/_integration/pyproject_reading.py
   https://github.com/pypa/setuptools_scm/blob/main/src/setuptools_scm/_config.py

.. py:data:: __all__
   :type: tuple[str, str]
   :value: ("ReadPyproject", "ReadPyprojectStrict")

   Module exports

.. py:data:: log
   :type: logging.Logger

   module level logger

"""

import abc
from pathlib import Path
from typing import NamedTuple

import setuptools_scm._log
from setuptools_scm._integration.toml import (
    TOML_RESULT,
    read_toml_content,
)

log = setuptools_scm._log.log.getChild("pyproject_reading")

__all__ = (
    "ReadPyproject",
    "ReadPyprojectStrict",
)


class PyProjectData(NamedTuple):
    path: Path
    tool_name: str
    project: TOML_RESULT
    section: TOML_RESULT

    @property
    def project_name(self):
        """Getter for project.name

        :returns: package name
        :rtype: str | None
        """
        return self.project.get("name")


class ReadPyprojectBase(abc.ABC):
    @abc.abstractmethod
    def update(self, d_target, d_other):
        """Subclass overload so can filter d_other

        :param d_target: parent dict
        :type d_target: dict[str, typing.Any]
        :param d_other: dict. Update parent
        :type d_other: dict[str, typing.Any]
        """
        ...

    def __call__(
        self,
        path=Path("pyproject.toml"),
        tool_name=["drain-swamp", "pipenv-unlock"],
        require_section=True,
    ):
        """Read in pyproject.toml and if necessary combine multiple sections into one

        Previously raised :py:exc:`FileNotFoundError` and :py:exc:`LookupError`
        instead will produce an empty dict

        :param path: Defaults to :code:`Path("pyproject.toml")` Absolute path to toml file
        :type path: pathlib.Path
        :param tool_name:

           Default ["drain-swamp", "pipenv-unlock"]. ``pyproject.toml``
           sections name or section name. First section name **MUST** be the package name

        :type tool_name: str | Sequence[str]
        :param require_section:

           Default True. This is ignored. Prevents extending package
           features and the possibility of pulling in multiple sections

        :type require_section: bool
        :returns: A convenience representation of ``pyproject.toml`` file
        :rtype: PyProjectData
        :raises:

           - :py:exc:`LookupError` -- Either toml file parsing failed
             or no such section

        """
        if isinstance(tool_name, str):
            seq_tool = (tool_name,)
        else:
            # assumes a Sequence[str]
            seq_tool = tool_name

        try:
            defn = read_toml_content(path)
        except FileNotFoundError:
            # None if require_section else {}
            defn = {}

        # Combine sections
        # Was preventing extending setuptools_scm
        d_section: TOML_RESULT = {}
        for tool_name in seq_tool:
            try:
                section = defn.get("tool", {})[tool_name]
            except (KeyError, LookupError) as e:
                error = f"{path} does not contain a tool.{tool_name} section"
                msg_warn = f"toml section missing {error!r}"
                # log.warning(msg_warn, exc_info=True)
                raise LookupError(msg_warn) from e
            else:
                # subclass must implement method, ``update``
                self.update(d_section, section)

        project = defn.get("project", {})

        # Use 1st tool name
        tool_name_2: str = seq_tool[0]

        return PyProjectData(path, tool_name_2, project, d_section)


class ReadPyproject(ReadPyprojectBase):
    def update(self, d_target, d_other):
        """
        :param d_target: parent dict
        :type d_target: dict[str, typing.Any]
        :param d_other: dict. Update parent
        :type d_other: dict[str, typing.Any]
        """
        d_target.update(d_other)


class ReadPyprojectStrict(ReadPyprojectBase):
    def update(self, d_target, d_other):
        """Confine to only setuptools_scm._config.Configuration data fields

        :param d_target: parent dict
        :type d_target: dict[str, typing.Any]
        :param d_other: dict. Update parent
        :type d_other: dict[str, typing.Any]
        """
        # import dataclasses
        # from setuptools_scm._config import Configuation
        # [field.name for field in fields(Configuration)]
        keys_allowed = (
            "relative_to",
            "root",
            "version_scheme",
            "local_scheme",
            "tag_regex",
            "parentdir_prefix_version",
            "fallback_version",
            "fallback_root",
            "write_to",
            "write_to_template",
            "version_file",
            "version_file_template",
            "parse",
            "git_describe_command",
            "dist_name",
            "version_cls",
            "search_parent_directories",
            "parent",
        )
        # dict comprehension to filter dict keys
        d_y = {k: v for k, v in d_other.items() if k in keys_allowed}
        d_target.update(d_y)



import re
import sys
from pathlib import Path

from packaging.version import parse
from sphinx_pyproject import SphinxConfig

from drain_swamp.pep518_read import find_project_root

path_docs = Path(__file__).parent
path_package_base = path_docs.parent
sys.path.insert(0, str(path_package_base))  # Needed ??

# @@@ editable
copyright = "2024–2024, Dave Faulkmore"
# The short X.Y.Z version.
version = "1.3.2"
# The full version, including alpha/beta/rc tags.
release = "1.3.2"
# The date of release, in "monthname day, year" format.
release_date = "August 11, 2024"
# @@@ end

v = parse(release)
version_short = f"{v.major}.{v.minor}"
# version_xyz = f"{v.major}.{v.minor}.{v.micro}"
version_xyz = version

# pyproject.toml search algo. Credit/Source: https://pypi.org/project/black/
srcs = (path_package_base,)
t_root = find_project_root(srcs)

try:
    from drain_swamp.constants_maybe import __version__ as version_semantic_full
except ImportError:
    # set during each tagged release
    version_long = version_xyz
else:
    # version file is generated during python -m build --sdist
    version_long = version_semantic_full

config = SphinxConfig(
    # Path(__file__).parent.parent.joinpath("pyproject.toml"),
    t_root[0] / "pyproject.toml",
    globalns=globals(),
    config_overrides={"version": version_long},
)

# This project is a fork from "Sphinx External ToC"
proj_project = config.name
proj_description = config.description
proj_authors = config.author

slug = re.sub(r"\W+", "-", proj_project.lower())
proj_master_doc = config.get("master_doc")
project = f"{proj_project} {version_xyz}"

###############
# Dynamic
###############
rst_epilog = """
.. |project_name| replace:: {slug}
.. |package-equals-release| replace:: drain_swamp=={release}
""".format(
    release=release, slug=slug
)

html_theme_options = {
    "description": proj_description,
    "show_relbars": True,
    "logo_name": False,
    "logo": "snip-logo.svg",
    "show_powered_by": False,
}

latex_documents = [
    (
        proj_master_doc,
        f"{slug}.tex",
        f"{proj_project} Documentation",
        proj_authors,
        "manual",  # manual, howto, jreport (Japanese)
        True,
    )
]
man_pages = [
    (
        proj_master_doc,
        slug,
        f"{proj_project} Documentation",
        [proj_authors],
        1,
    )
]
texinfo_documents = [
    (
        proj_master_doc,
        slug,
        f"{proj_project} Documentation",
        proj_authors,
        slug,
        proj_description,
        "Miscellaneous",
    )
]

#################
# Static
#################
ADDITIONAL_PREAMBLE = r"""
\DeclareUnicodeCharacter{20BF}{\'k}
"""

latex_elements = {
    "sphinxsetup": "verbatimforcewraps",
    "extraclassoptions": "openany,oneside",
    "preamble": ADDITIONAL_PREAMBLE,
}

html_sidebars = {
    "**": [
        "about.html",
        "searchbox.html",
        "navigation.html",
        "relations.html",
    ],
}

intersphinx_mapping = {
    "python": (  # source https://docs.python.org/3/objects.inv
        "https://docs.python.org/3",
        ("objects-python.inv", None),
    ),
    "snip": (
        "https://drain-swamp.readthedocs.io/en/latest",
        ("objects-snip.inv", "objects-snip.txt"),
    ),
    "black": (
        "https://github.com/psf/black",
        ("objects-black.inv", "objects-black.txt"),
    ),
    "python-missing": (
        "https://github.com/python/cpython/blob",
        ("objects-python-missing.inv", "objects-python-missing.txt"),
    ),
    "setuptools-scm": (
        "https://setuptools-scm.readthedocs.io/en/stable",
        ("objects-setuptools-scm.inv", "objects-setuptools-scm.txt"),
    ),
}
intersphinx_disabled_reftypes = ["std:doc"]

extlinks = {
    "pypi_org": (  # url to: aiologger
        "https://pypi.org/project/%s",
        "%s",
    ),
}

# spoof user agent to prevent broken links
# curl -A "Mozilla/5.0 (X11; Ubuntu; Linux i686; rv:24.0) Gecko/20100101 Firefox/24.0" --head "https://github.com/python/cpython/blob/3.12/Lib/unittest/case.py#L193"
linkcheck_request_headers = {
    "https://github.com/": {
        "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux i686; rv:24.0) Gecko/20100101 Firefox/24.0",
    },
    "https://docs.github.com/": {
        "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux i686; rv:24.0) Gecko/20100101 Firefox/24.0",
    },
}

# Ignore unfixable WARNINGS
# in pyproject.toml --> nitpicky = true
# in conf.py --> nitpicky = True
nitpick_ignore = [
    ("py:class", "ValidatorType"),
]



from collections.abc import Sequence
from functools import lru_cache
from pathlib import Path
from typing import Any

__all__ = (
    "find_project_root",
    "find_pyproject_toml",
)

def _is_ok(test: Any | None) -> bool: ...
@lru_cache
def find_project_root(
    srcs: Sequence[Any] | None,
    stdin_filename: str | None = None,
) -> tuple[Path, str]: ...
def find_pyproject_toml(
    path_search_start: tuple[str, ...],
    stdin_filename: str | None = None,
) -> str | None: ...



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

..

Unittest for entrypoint, cli_unlock

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_cli_unlock.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_cli_unlock.py

Needs a config file to specify exact files to include / omit from report.
Will fail with exit code 1 even with 100% coverage

.. seealso::

   https://github.com/pytest-dev/pytest-cov/issues/373#issuecomment-1472861775

"""

import logging
import logging.config
import traceback
from pathlib import Path
from unittest.mock import patch

import pytest
from click.testing import CliRunner

from drain_swamp.backend_abc import BackendType
from drain_swamp.backend_setuptools import BackendSetupTools  # noqa: F401
from drain_swamp.cli_unlock import (
    create_links,
    dependencies_lock,
    dependencies_unlock,
    entrypoint_name,
    main,
    state_is_lock,
)
from drain_swamp.constants import (
    LOGGING,
    g_app_name,
)
from drain_swamp.parser_in import TomlParser


def test_cli_main():
    """Minimally test package version is printed"""
    runner = CliRunner()
    # --version
    """
    cmd = ["--version"]
    result = runner.invoke(main, cmd)
    assert result.exit_code == 0
    assert "version" in result.stdout
    """

    # --help
    cmd = ["--help"]
    result = runner.invoke(main, cmd)
    assert result.exit_code == 0
    assert f"Command-line for {entrypoint_name}. Prints usage" in result.stdout


testdata_lock_unlock_successfully = (
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "little_shop_of_horrors_shrine_candles",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
    ),
)
ids_lock_unlock_successfully = ("With optional ci/kit.in and additional folder ci",)


@pytest.mark.parametrize(
    "path_pyproject_toml, id_, additional_folders, additional_files",
    testdata_lock_unlock_successfully,
    ids=ids_lock_unlock_successfully,
)
def test_lock_unlock_successfully(
    path_pyproject_toml,
    id_,
    additional_folders,
    additional_files,
    tmp_path,
    caplog,
    has_logging_occurred,
    prep_pyproject_toml,
    prep_cmd_unlock_lock,
    prepare_files_empties,
):
    # pytest --showlocals --log-level INFO -k "test_lock_unlock_successfully" -v tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        path_f = prep_pyproject_toml(path_pyproject_toml, path_tmp_dir)
        tp = TomlParser(path_f)
        d_pyproject_toml = tp.d_pyproject_toml
        prepare_files_empties(
            d_pyproject_toml,
            path_tmp_dir,
            d_add_files=additional_files,
        )
        cmd = prep_cmd_unlock_lock(
            path_tmp_dir,
            snip_co=id_,
            add_folders=additional_folders,
        )
        is_locked = BackendType.is_locked(path_f)
        expected = path_f.read_text()

        # act
        if is_locked:
            result = runner.invoke(dependencies_unlock, cmd)
            assert result.exit_code == 0
            result = runner.invoke(dependencies_lock, cmd)
            assert result.exit_code == 0
            actual = path_f.read_text()
        else:
            result = runner.invoke(dependencies_lock, cmd)
            assert result.exit_code == 0
            result = runner.invoke(dependencies_unlock, cmd)
            assert result.exit_code == 0
            actual = path_f.read_text()

        assert has_logging_occurred(caplog)
        assert expected == actual


testdata_lock_unlock_and_back_wo_prepare = (
    (
        dependencies_unlock,
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "little_shop_of_horrors_shrine_candles",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        3,  # PyProjectTOMLReadError
    ),
    (
        dependencies_lock,
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "little_shop_of_horrors_shrine_candles",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        3,  # PyProjectTOMLReadError
    ),
)
ids_lock_unlock_and_back_wo_prepare = (
    "call dependencies_unlock. Additional file ci/kit.in",
    "call dependencies_lock. Additional file ci/kit.in",
)


@pytest.mark.parametrize(
    "func, path_pyproject_toml, id_, additional_folders, additional_files, expected_exit_code",
    testdata_lock_unlock_and_back_wo_prepare,
    ids=ids_lock_unlock_and_back_wo_prepare,
)
def test_lock_unlock_and_back_wo_prepare(
    func,
    path_pyproject_toml,
    id_,
    additional_folders,
    additional_files,
    expected_exit_code,
    tmp_path,
    prep_pyproject_toml,
    prep_cmd_unlock_lock,
    prepare_files_empties,
    caplog,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_lock_unlock_and_back_wo_prepare" -v tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        # path_f = prep_pyproject_toml(path_pyproject_toml, path_tmp_dir)
        cmd = prep_cmd_unlock_lock(
            path_tmp_dir,
            add_folders=additional_folders,
            snip_co=id_,
        )
        # Call cli func blind; no BackendType.is_locked
        result = runner.invoke(func, cmd)

        logger.info(result.output)
        tb = result.exc_info[2]
        # msg_info = f"traceback: {pprint(traceback.format_tb(tb))}"
        msg_info = f"traceback: {traceback.format_tb(tb)}"
        logger.info(msg_info)

        actual_exit_code = result.exit_code
        assert actual_exit_code == expected_exit_code


testdata_lock_unlock_and_back_with_prepare = (
    (
        dependencies_lock,
        Path(__file__).parent.joinpath(
            "_good_files", "backend-unsupported.pyproject_toml"
        ),
        "little_shop_of_horrors_shrine_candles",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        True,
        True,
        5,  # BackendNotSupportedError
    ),
    (
        dependencies_unlock,
        Path(__file__).parent.joinpath(
            "_good_files", "backend-unsupported.pyproject_toml"
        ),
        "little_shop_of_horrors_shrine_candles",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        True,
        True,
        5,  # BackendNotSupportedError
    ),
    (
        dependencies_lock,
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "little_shop_of_horrors_shrine_candles",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        True,
        False,
        6,  # MissingRequirementsFoldersFiles
    ),
    (
        dependencies_unlock,
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "little_shop_of_horrors_shrine_candles",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        True,
        False,
        6,  # MissingRequirementsFoldersFiles
    ),
    (
        dependencies_lock,
        Path(__file__).parent.joinpath("_bad_files", "snippet-nested.pyproject_toml"),
        "little_shop_of_horrors_shrine_candles",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        True,
        True,
        8,
    ),
    (
        dependencies_unlock,
        Path(__file__).parent.joinpath("_bad_files", "snippet-nested.pyproject_toml"),
        "little_shop_of_horrors_shrine_candles",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        True,
        True,
        8,
    ),
    (
        dependencies_lock,
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "little_shop_of_horrors_obedient_girl_friend",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        True,
        True,
        9,
    ),
    (
        dependencies_unlock,
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "little_shop_of_horrors_obedient_girl_friend",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        True,
        True,
        9,
    ),
)
ids_lock_unlock_and_back_with_prepare = (
    "lock unsupported backend",
    "unlock unsupported backend",
    "lock missing folders and files",
    "unlock missing folders and files",
    "lock Snippet is invalid",
    "unlock Snippet is invalid",
    "lock Snippet no match",
    "unlock Snippet no match",
)


@pytest.mark.parametrize(
    "func, path_pyproject_toml, id_, additional_folders, additional_files, is_prep_pyproject_toml, is_prep_files, expected_exit_code",
    testdata_lock_unlock_and_back_with_prepare,
    ids=ids_lock_unlock_and_back_with_prepare,
)
def test_lock_unlock_and_back_with_prepare(
    func,
    path_pyproject_toml,
    id_,
    additional_folders,
    additional_files,
    is_prep_pyproject_toml,
    is_prep_files,
    expected_exit_code,
    tmp_path,
    caplog,
    has_logging_occurred,
    prep_pyproject_toml,
    prep_cmd_unlock_lock,
    prepare_files_empties,
):
    # pytest --showlocals --log-level INFO -k "test_lock_unlock_and_back_with_prepare" -v tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        if is_prep_pyproject_toml:
            path_f = prep_pyproject_toml(path_pyproject_toml, path_tmp_dir)
            tp = TomlParser(path_f)
            d_pyproject_toml = tp.d_pyproject_toml
            if is_prep_files:
                prepare_files_empties(
                    d_pyproject_toml,
                    path_tmp_dir,
                    d_add_files=additional_files,
                )

        cmd = prep_cmd_unlock_lock(
            path_tmp_dir,
            snip_co=id_,
            add_folders=additional_folders,
        )

        # Call cli func blind; no BackendType.is_locked
        result = runner.invoke(func, cmd)

        logger.info(f"exit_code: {result.exit_code}")
        logger.info(f"exception: {result.exception}")
        logger.info(f"output: {result.output}")

        tb = result.exc_info[2]
        # msg_info = f"traceback: {pprint(traceback.format_tb(tb))}"
        msg_info = f"traceback: {traceback.format_tb(tb)}"
        logger.info(msg_info)

        actual_exit_code = result.exit_code
        assert actual_exit_code == expected_exit_code


testdata_lock_unlock_and_back_with_patch = (
    (
        dependencies_lock,
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "little_shop_of_horrors_shrine_candles",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        True,
        True,
        7,  # BackendNotSupportedError
    ),
)
ids_lock_unlock_and_back_with_patch = (
    "pip-tools not installed. pip-compile command unavailable",
)


@pytest.mark.parametrize(
    "func, path_pyproject_toml, id_, additional_folders, additional_files, is_prep_pyproject_toml, is_prep_files, expected_exit_code",
    testdata_lock_unlock_and_back_with_patch,
    ids=ids_lock_unlock_and_back_with_patch,
)
def test_lock_unlock_and_back_with_patch(
    func,
    path_pyproject_toml,
    id_,
    additional_folders,
    additional_files,
    is_prep_pyproject_toml,
    is_prep_files,
    expected_exit_code,
    tmp_path,
    caplog,
    has_logging_occurred,
    prep_pyproject_toml,
    prep_cmd_unlock_lock,
    prepare_files_empties,
):
    # pytest --showlocals --log-level INFO -k "test_lock_unlock_and_back_with_patch" -v tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        if is_prep_pyproject_toml:
            path_f = prep_pyproject_toml(path_pyproject_toml, path_tmp_dir)
            tp = TomlParser(path_f)
            d_pyproject_toml = tp.d_pyproject_toml
            if is_prep_files:
                prepare_files_empties(
                    d_pyproject_toml,
                    path_tmp_dir,
                    d_add_files=additional_files,
                )

        cmd = prep_cmd_unlock_lock(
            path_tmp_dir,
            snip_co=id_,
            add_folders=additional_folders,
        )

        # Call cli func blind; no BackendType.is_locked
        with patch(f"{g_app_name}.lock_toggle.is_piptools", return_value=False):
            result = runner.invoke(func, cmd)

        logger.info(result.output)
        tb = result.exc_info[2]
        # msg_info = f"traceback: {pprint(traceback.format_tb(tb))}"
        msg_info = f"traceback: {traceback.format_tb(tb)}"
        logger.info(msg_info)

        actual_exit_code = result.exit_code
        assert actual_exit_code == expected_exit_code


testdata_lock_unlock_and_back_card_monte = (
    (
        dependencies_lock,
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        Path(__file__).parent.joinpath("_bad_files", "backend_only.pyproject_toml"),
        "little_shop_of_horrors_shrine_candles",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        True,
        True,
        4,  # PyProjectTOMLParseError
    ),
    (
        dependencies_unlock,
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        Path(__file__).parent.joinpath("_bad_files", "backend_only.pyproject_toml"),
        "little_shop_of_horrors_shrine_candles",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        True,
        True,
        4,  # PyProjectTOMLParseError
    ),
)
ids_lock_unlock_and_back_card_monte = (
    "lock. Cause parsing to fail, but not during prep",
    "unlock. Cause parsing to fail, but not during prep",
)


@pytest.mark.parametrize(
    "func, path_pyproject_toml_prep, path_pyproject_toml_test, id_, additional_folders, additional_files, is_prep_pyproject_toml, is_prep_files, expected_exit_code",
    testdata_lock_unlock_and_back_card_monte,
    ids=ids_lock_unlock_and_back_card_monte,
)
def test_lock_unlock_and_back_card_monte(
    func,
    path_pyproject_toml_prep,
    path_pyproject_toml_test,
    id_,
    additional_folders,
    additional_files,
    is_prep_pyproject_toml,
    is_prep_files,
    expected_exit_code,
    tmp_path,
    caplog,
    has_logging_occurred,
    prep_pyproject_toml,
    prep_cmd_unlock_lock,
    prepare_files_empties,
):
    """Would like to cause PyProjectTOMLParseError; exit code 4. But it's
    a catch 22. To prepare the files, need a viable pyproject.toml. To
    trigger the exception need a non-viable pyproject.toml

    So we are gonna play a magic trick. Here you see it; now you don't.
    """
    # pytest --showlocals --log-level INFO -k "test_lock_unlock_and_back_card_monte" -v tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        # prepare -- use viable pyproject.toml
        path_tmp_dir = Path(tmp_dir_path)
        if is_prep_pyproject_toml:
            path_f_prep = prep_pyproject_toml(path_pyproject_toml_prep, path_tmp_dir)
            tp = TomlParser(path_f_prep)
            d_pyproject_toml = tp.d_pyproject_toml
            if is_prep_files:
                prepare_files_empties(
                    d_pyproject_toml,
                    path_tmp_dir,
                    d_add_files=additional_files,
                )
            prep_pyproject_toml(path_pyproject_toml_test, path_tmp_dir)

        cmd = prep_cmd_unlock_lock(
            path_tmp_dir,
            snip_co=id_,
            add_folders=additional_folders,
        )

        result = runner.invoke(func, cmd)

        logger.info(result.output)
        tb = result.exc_info[2]
        # msg_info = f"traceback: {pprint(traceback.format_tb(tb))}"
        msg_info = f"traceback: {traceback.format_tb(tb)}"
        logger.info(msg_info)

        actual_exit_code = result.exit_code
        assert actual_exit_code == expected_exit_code


testdata_lock_unlock_and_back_optionals = (
    (
        dependencies_lock,
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "little_shop_of_horrors_shrine_candles",
        (Path("ci"),),
        {"ci": Path("ci/kit.in")},
        {
            "web": Path("ci/web.in"),
            "pytest": Path("ci/pytest.in"),
        },
        9,
    ),
    (
        dependencies_unlock,
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "little_shop_of_horrors_shrine_candles",
        ("ci",),
        {"ci": Path("ci/kit.in")},
        {
            "web": Path("ci/web.in"),
            "pytest": Path("ci/pytest.in"),
        },
        9,
    ),
)
ids_lock_unlock_and_back_optionals = (
    "lock create two optionals",
    "unlock create two optionals ",
)


@pytest.mark.parametrize(
    "func, path_pyproject_toml, id_, additional_folders, additional_files, d_optionals, expected_count",
    testdata_lock_unlock_and_back_optionals,
    ids=ids_lock_unlock_and_back_optionals,
)
def test_lock_unlock_and_back_optionals(
    func,
    path_pyproject_toml,
    id_,
    additional_folders,
    additional_files,
    d_optionals,
    expected_count,
    tmp_path,
    caplog,
    has_logging_occurred,
    prep_pyproject_toml,
    prep_cmd_unlock_lock,
    prepare_files_empties,
):
    # pytest --showlocals --log-level INFO -k "test_lock_unlock_and_back_optionals" -v tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        path_f = prep_pyproject_toml(path_pyproject_toml, path_tmp_dir)
        tp = TomlParser(path_f)
        d_pyproject_toml = tp.d_pyproject_toml
        prepare_files_empties(
            d_pyproject_toml,
            path_tmp_dir,
            d_add_files=additional_files,
            d_optionals=d_optionals,
        )

        cmd = prep_cmd_unlock_lock(
            path_tmp_dir,
            d_opts=d_optionals,
            add_folders=additional_folders,
            snip_co=id_,
        )

        result = runner.invoke(func, cmd)
        # logger.info(f"exception: {result.exception}")
        # logger.info(f"output: {result.output}")
        assert result.exit_code == 0

        inst = BackendType.load_factory(
            path_tmp_dir,
            optionals=d_optionals,
            additional_folders=additional_folders,
        )
        gen = inst.in_files()
        files = list(gen)
        actual_count = len(files)

        assert has_logging_occurred(caplog)

        assert actual_count == expected_count


testdata_is_lock = (
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        0,
    ),
    (
        Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        ),
        1,
    ),
    (
        Path(__file__).parent.joinpath(
            "_bad_files", "static_dependencies.pyproject_toml"
        ),
        4,
    ),
    (
        Path(__file__).parent.joinpath("_changelog_files", "CHANGES-empty.rst"),
        3,
    ),
)
ids_is_lock = (
    "locked",
    "unlocked",
    "static dependencies. No tool.setuptools.dynamic section",
    "not a toml file",
)


@pytest.mark.parametrize(
    "path_pyproject_toml, expected_exit_code",
    testdata_is_lock,
    ids=ids_is_lock,
)
def test_state_is_lock(
    path_pyproject_toml,
    expected_exit_code,
    tmp_path,
    prep_pyproject_toml,
    caplog,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_state_is_lock" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        path_f = prep_pyproject_toml(path_pyproject_toml, path_tmp_dir)
        # click chokes! Attempts to Path.resolve a float
        invalids = (1.2345,)
        for invalid in invalids:
            cmd = [
                "--path",
                invalid,
            ]
            result = runner.invoke(state_is_lock, cmd)
            assert result.exception.__class__ == TypeError
            # logger.info(f"result.output: {result.output}")
            # logger.info(f"result.exit_code: {result.exit_code}")
            # tb = result.exc_info[2]
            # logger.info(f"traceback: {traceback.format_tb(tb)}")
            pass

        relpath_f = path_f.relative_to(path_tmp_dir)

        cmds = [
            ("--path", str(relpath_f)),  # relative path
            (
                "--path",
                path_f,
            ),  # absolute path
            (
                "--path",
                "",
            ),  # --> default path
        ]

        for func_cmd in cmds:
            result = runner.invoke(state_is_lock, func_cmd)
            logger.info(f"result.exception: {result.exception}")
            logger.info(f"result.output: {result.output}")
            assert result.exit_code == expected_exit_code

        # assert has_logging_occurred(caplog)
        pass


def test_create_links_exceptions(
    caplog, tmp_path, prep_pyproject_toml, has_logging_occurred, prepare_folders_files
):
    # pytest --showlocals --log-level INFO -k "test_create_links_exceptions" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        # PyProjectTOMLReadError (3) -- note BackendType.load_factory call is wrong
        func_cmd = [
            "--path",
            tmp_dir_path,
        ]
        result = runner.invoke(create_links, func_cmd)
        assert result.exit_code == 3

        # malformed toml / backend only --> PyProjectTOMLParseError
        path_pyproject_toml_4 = Path(__file__).parent.joinpath(
            "_bad_files", "backend_only.pyproject_toml"
        )
        expected = 4
        path_f = prep_pyproject_toml(path_pyproject_toml_4, path_tmp_dir)
        result = runner.invoke(create_links, func_cmd)
        assert result.exit_code == expected

        # unsupported backend (5) -- note BackendType.load_factory call is wrong
        path_pyproject_toml_5 = Path(__file__).parent.joinpath(
            "_good_files", "backend-unsupported.pyproject_toml"
        )
        expected = 5
        #    prepare
        path_f = prep_pyproject_toml(path_pyproject_toml_5, path_tmp_dir)
        result = runner.invoke(create_links, func_cmd)
        assert result.exit_code == expected

        # Path is expected to be a folder, not a file (2)
        path_pyproject_toml_7 = Path(__file__).parent.joinpath(
            "_bad_files", "static_dependencies.pyproject_toml"
        )
        expected = 2
        #    prepare
        path_f = prep_pyproject_toml(path_pyproject_toml_7, path_tmp_dir)
        func_cmd = [
            "--path",
            str(path_f),
        ]
        inst = BackendType.load_factory(
            path_f,
            parent_dir=tmp_dir_path,
        )
        assert inst.parent_dir == path_tmp_dir
        with patch(
            f"{g_app_name}.cli_unlock.BackendType.load_factory",
            return_value=inst,
        ):
            result = runner.invoke(create_links, func_cmd)
            assert result.exit_code == expected

        """static dependencies (7)

        - No dynamic section in pyproject.toml
        - no ``dependencies`` key

        unlocked == "0" locked == "1". None use current lock state
        """
        expected = 7
        func_cmd = [
            "--path",
            tmp_dir_path,
        ]
        with patch(
            f"{g_app_name}.cli_unlock.BackendType.load_factory",
            return_value=inst,
        ):
            result = runner.invoke(create_links, func_cmd)
            assert result.exit_code == expected

            """
            logger.info(f"result.output: {result.output}")
            logger.info(f"result.exit_code: {result.exit_code}")
            logger.info(f"result.exception: {result.exception}")
            tb = result.exc_info[2]
            logger.info(f"traceback: {traceback.format_tb(tb)}")
            assert has_logging_occurred(caplog)


            assert result.stdout.rstrip() == expected_text

            expected_text = "In pyproject.toml no section, tool.setuptools.dynamic"
            """
            pass


testdata_create_links_set_lock = (
    (
        Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        ),
        (
            "requirements/prod.in",
            "requirements/pip.in",
            "requirements/manage.in",
        ),
        (
            "requirements/prod.unlock",
            "requirements/pip.unlock",
            "requirements/manage.unlock",
        ),
        (
            "requirements/prod.lock",
            "requirements/pip.lock",
            "requirements/manage.lock",
        ),
        "1",
        0,
    ),
    (
        Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        ),
        (
            "requirements/prod.in",
            "requirements/pip.in",
            "requirements/manage.in",
        ),
        (
            "requirements/prod.unlock",
            "requirements/pip.unlock",
            "requirements/manage.unlock",
        ),
        (
            "requirements/prod.lock",
            "requirements/pip.lock",
            "requirements/manage.lock",
        ),
        "0",
        0,
    ),
    (
        Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        ),
        (
            "requirements/prod.in",
            "requirements/pip.in",
            "requirements/manage.in",
        ),
        (
            "requirements/prod.unlock",
            "requirements/pip.unlock",
            "requirements/manage.unlock",
        ),
        (
            "requirements/prod.lock",
            "requirements/pip.lock",
            "requirements/manage.lock",
        ),
        None,
        0,
    ),
)
ids_create_links_set_lock = (
    "set symlink to dependency locked",
    "set symlink to dependency unlocked",
    "set symlink to current lock state",
)


@pytest.mark.parametrize(
    "path_pyproject_toml, seq_in, seq_unlock, seq_lock, set_lock, expected",
    testdata_create_links_set_lock,
    ids=ids_create_links_set_lock,
)
def test_create_links_set_lock(
    path_pyproject_toml,
    seq_in,
    seq_unlock,
    seq_lock,
    set_lock,
    expected,
    caplog,
    tmp_path,
    prep_pyproject_toml,
    has_logging_occurred,
    prepare_folders_files,
):
    # pytest --showlocals --log-level INFO -k "test_create_links_set_lock" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        # prepare
        path_config = prep_pyproject_toml(path_pyproject_toml, path_tmp_dir)

        #    .in
        prepare_folders_files(seq_in, path_tmp_dir)
        #    .unlock .lock
        prepare_folders_files(seq_unlock, path_tmp_dir)
        prepare_folders_files(seq_lock, path_tmp_dir)

        inst = BackendType.load_factory(
            path_config,
            parent_dir=tmp_dir_path,
        )
        for in_ in inst.in_files():
            assert in_.exists()

        #    cmd
        click_true = ("1", "true", "t", "yes", "y", "on")
        click_false = ("0", "false", "f", "no", "n", "off")
        if set_lock in click_true:
            from_suffix = ".lock"
        elif set_lock == click_false:
            from_suffix = ".unlock"
        else:
            is_locked = inst.is_locked(path_config)
            if is_locked:
                from_suffix = ".lock"
            else:
                from_suffix = ".unlock"

        #    Defaults to None, but can't pass in None explicitly
        func_cmd = [
            "--path",
            tmp_dir_path,
        ]
        if set_lock is not None and isinstance(set_lock, str):
            func_cmd.extend(["--set-lock", set_lock])

        with patch(
            f"{g_app_name}.cli_unlock.BackendType.load_factory",
            return_value=inst,
        ):
            result = runner.invoke(create_links, func_cmd)

        #    prove symlinks created
        for unlock_relpath in seq_in:
            lnk_relpath = unlock_relpath.replace(".in", ".lnk")
            path_lnk = path_tmp_dir.joinpath(lnk_relpath)
            is_symlink = path_lnk.is_symlink()
            is_suffix_match = path_lnk.resolve().suffix == from_suffix
            assert (
                is_suffix_match
            ), f"{path_lnk} does not resolve to a {from_suffix} file"
            # assert has_logging_occurred(caplog)
            assert is_symlink

        assert result.exit_code == expected


@pytest.mark.parametrize(
    "path_pyproject_toml, seq_in, seq_unlock, seq_lock, set_lock, expected",
    testdata_create_links_set_lock,
    ids=ids_create_links_set_lock,
)
def test_create_links_missing_files(
    path_pyproject_toml,
    seq_in,
    seq_unlock,
    seq_lock,
    set_lock,
    expected,
    caplog,
    tmp_path,
    prep_pyproject_toml,
    has_logging_occurred,
    prepare_folders_files,
):
    # pytest --showlocals --log-level INFO -k "test_create_links_missing_files" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        # prepare
        path_config = prep_pyproject_toml(path_pyproject_toml, path_tmp_dir)

        #    .in
        prepare_folders_files(seq_in, path_tmp_dir)
        #    try without .unlock and .lock files --> MissingRequirementsFoldersFiles(6)
        inst = BackendType.load_factory(
            path_config,
            parent_dir=tmp_dir_path,
        )
        for in_ in inst.in_files():
            assert in_.exists()

        #    cmd
        #    Defaults to None, but can't pass in None explicitly
        func_cmd = [
            "--path",
            tmp_dir_path,
        ]
        if set_lock is not None and isinstance(set_lock, str):
            func_cmd.extend(["--set-lock", set_lock])

        expected_exit_code = 6
        with patch(
            f"{g_app_name}.cli_unlock.BackendType.load_factory",
            return_value=inst,
        ):
            result = runner.invoke(create_links, func_cmd)
            actual_exit_code = result.exit_code
            assert actual_exit_code == expected_exit_code



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_patch_pyproject_reading.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_patch_pyproject_reading.py

"""

from pathlib import (
    Path,
    PurePath,
)

import pytest

from drain_swamp.monkey.patch_pyproject_reading import (
    PyProjectData,
    ReadPyproject,
)

testdata_pyproject_data = (
    (
        "hope",
        {},
        {"name": "bob"},
        "bob",
    ),
)
ids_pyproject_data = ("empty section, pyproject.toml contains only name",)


@pytest.mark.parametrize(
    "tool_name, section, project, expected",
    testdata_pyproject_data,
    ids=ids_pyproject_data,
)
def test_pyproject_data(tool_name, section, expected, project, tmp_path):
    "From pyproject.toml confirm can get project name"
    data = PyProjectData(tmp_path, tool_name, project, section)
    assert data.project_name == expected


testdata_read_pyproject = (
    (
        Path(__file__).parent.joinpath(
            "_good_files",
            "complete-manage-pip-prod-unlock.pyproject_toml",
        ),
        "drain-swamp",
        "drain-swamp",
    ),
    (
        Path(__file__).parent.joinpath(
            "_good_files",
            "complete-manage-pip-prod-unlock.pyproject_toml",
        ),
        ("drain-swamp", "pipenv-unlock"),
        "drain-swamp",
    ),
)
ids_read_pyproject = (
    "tool_name str",
    "tool_name Sequence[str]",
)


@pytest.mark.parametrize(
    "p_toml_file, tool_name, expected_tool_name",
    testdata_read_pyproject,
    ids=ids_read_pyproject,
)
def test_read_pyproject(
    p_toml_file, tool_name, expected_tool_name, tmp_path, prep_pyproject_toml
):
    # prepare
    prep_pyproject_toml(p_toml_file, tmp_path)

    data = ReadPyproject()(path=p_toml_file, tool_name=tool_name)
    assert issubclass(type(data.path), PurePath)
    assert isinstance(data.tool_name, str)
    assert isinstance(data.section, dict)
    assert isinstance(data.project, dict)
    assert data.path == p_toml_file
    assert data.tool_name == expected_tool_name


def test_read_pyproject_exception(tmp_path):
    # No pyproject.toml file and no such section
    tool_name = "Bob"
    p_toml_file = tmp_path.joinpath("pyproject.toml")
    with pytest.raises(LookupError):
        ReadPyproject()(path=p_toml_file, tool_name=tool_name)

    # no [tool.Bob] section
    p_toml_file = Path(__file__).parent.joinpath(
        "_good_files",
        "complete-manage-pip-prod-unlock.pyproject_toml",
    )
    with pytest.raises(LookupError):
        ReadPyproject()(path=p_toml_file, tool_name=tool_name)

    # str
    tool_name = "pipenv-unlock"
    data_0 = ReadPyproject()(path=p_toml_file, tool_name=tool_name)
    assert isinstance(data_0.project, dict)
    assert isinstance(data_0.section, dict)
    assert data_0.project_name == "complete-awesome-perfect"
    #    from section drain-swamp
    assert "copyright_start_year" not in data_0.section.keys()
    #    from section pipenv-unlock
    assert "version_file" in data_0.section.keys()

    # # Sequence[str]
    tool_name = ["drain-swamp", "pipenv-unlock"]
    data_1 = ReadPyproject()(path=p_toml_file, tool_name=tool_name)
    assert isinstance(data_1.project, dict)
    assert isinstance(data_1.section, dict)
    assert data_1.project_name == "complete-awesome-perfect"
    #    from section drain-swamp
    assert "copyright_start_year" in data_1.section.keys()
    #    from section pipenv-unlock
    assert "version_file" in data_1.section.keys()



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Get scm version, limit to setuptool-scm codebase

"""

import sys
from pathlib import Path

import click

# pep366 ...
# https://stackoverflow.com/a/34155199
if __name__ == "__main__" and __spec__ is None:  # pragma: no cover
    # Package not installed
    # python src/drain_swamp/cli_scm_version.py get
    import importlib.util

    path_d = Path(__file__).parent
    rev_mods = []
    while path_d.joinpath("__init__.py").exists():
        name = path_d.name
        path_prev = path_d
        path_d = path_d.parent
        rev_mods.append(name)
    # One level above top package --> sys.path.insert
    sys.path.insert(1, str(path_d))
    # parent (aka package) dotted path
    dotted_path = ".".join(reversed(rev_mods))

    # print(f"str(path_d): {str(path_d)}", file=sys.stderr)
    # print(f"dotted_path: {dotted_path}", file=sys.stderr)
    # print(f"path_prev: {path_prev}", file=sys.stderr)
    pass

    # import top package level
    path_top = path_prev.joinpath("__init__.py")
    spec_top = importlib.util.spec_from_file_location(name, path_top)
    mod_top = importlib.util.module_from_spec(spec_top)
    sys.modules[dotted_path] = mod_top
    spec_top.loader.exec_module(mod_top)

    # __spec__ is None. Set __spec__ rather than :code:`__package__ = dotted_path`
    dotted_path_this = f"{dotted_path}.{Path(__file__).stem}"
    spec_this = importlib.util.spec_from_file_location(dotted_path_this, Path(__file__))
    __spec__ = spec_this
else:  # pragma: no cover
    pass

from drain_swamp.monkey.wrap_get_version import (
    scm_version,
    write_to_file,
)

entrypoint_name = "scm-version"  # noqa: F401

help_path = "The root directory [default: pyproject.toml directory]"
help_is_write = "Write flag. 1 write, 0 not write"
help_write_to = "Override version file relative path"

EPILOG_SCM_VERSION_GET = """
EXIT CODES

3 -- missing pyproject.toml or missing sections tool.drain-swamp and
     tool.pipenv-unlock

4 -- invalid semantic str. Skip. Do not write that nonsense to version file

"""
EPILOG_SCM_VERSION_WRITE = """
EXIT CODES

3 -- missing pyproject.toml or missing sections tool.drain-swamp and
     tool.pipenv-unlock

4 -- invalid semantic str. Skip. Do not write that nonsense to version file

"""


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
)
def main():
    """Command-line for scm-version. Prints usage"""


@main.command(
    "get",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_SCM_VERSION_GET,
)
@click.option(
    "-p",
    "--path",
    default=Path.cwd(),
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True),
    help=help_path,
)
@click.option(
    "--is-write",
    "is_write",
    default=False,
    help=help_is_write,
    is_flag=True,
)
@click.option(
    "-w",
    "--write-to",
    "write_to",
    default=None,
    type=click.STRING,
    help=help_write_to,
)
def get_scm_version(path, is_write, write_to):
    """scm version, setuptools-scm codebase only

    \f
    :param path:

       The root directory [default: pyproject.toml directory]

    :type path: pathlib.Path
    :param is_write: Default False. True to write to file otherwise don't write
    :type is_write: bool
    :param write_to: During testing provide alternative version file path
    :type write_to: str | None
    """
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path_cwd = Path(path)
    else:  # pragma: no cover
        path_cwd = path

    # Both scm_version and write_to_file expect relative_to not package base folder
    path_relative_to = path_cwd.joinpath("pyproject.toml")
    relative_to_path = str(path_relative_to)

    str_scm_ver = scm_version(relative_to_path)

    click.echo(str_scm_ver)

    if is_write:
        try:
            write_to_file(relative_to_path, str_scm_ver, write_to=write_to)
        except LookupError as e:
            msg_warn = str(e)
            exit_code = 3
        except ValueError as e:
            # invalid semantic version str (from git?!)
            msg_warn = str(e)
            exit_code = 4
        else:
            exit_code = 0

        if exit_code != 0:
            click.secho(msg_warn, fg="red", err=True)
            sys.exit(exit_code)
        else:  # pragma: no cover
            pass
    else:  # pragma: no cover
        pass


@main.command(
    "write",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_SCM_VERSION_WRITE,
)
@click.argument("scm_ver")
@click.option(
    "-p",
    "--path",
    default=Path.cwd(),
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True),
    help=help_path,
)
@click.option(
    "-w",
    "--write-to",
    "write_to",
    default=None,
    type=click.STRING,
    help=help_write_to,
)
def write_scm_version(scm_ver, path, write_to):
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path_cwd = Path(path)
    else:  # pragma: no cover
        path_cwd = path

    # Both scm_version and write_to_file expect relative_to not package base folder
    path_relative_to = path_cwd.joinpath("pyproject.toml")
    relative_to_path = str(path_relative_to)

    # Possible for there to be no pyproject.toml or missing sections
    try:
        write_to_file(relative_to_path, scm_ver, write_to=write_to)
    except LookupError as e:
        # issue with pyproject.toml
        msg_warn = str(e)
        exit_code = 3
    except ValueError as e:
        # invalid semantic version str
        msg_warn = str(e)
        exit_code = 4
    else:
        exit_code = 0

    if exit_code != 0:
        # Write version file skipped
        click.secho(msg_warn, fg="red", err=True)
    else:  # pragma: no cover
        pass
    sys.exit(exit_code)


if __name__ == "__main__":  # pragma: no cover
    main()



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Generate the .lnk files. Do not change the lock/unlock state

Usage

.. code-block:: shell

   python -m build -C--set-lock=0 --sdist
   python -m build -C--set-lock=0 -C--kind="tag" --sdist
   python -m build -C--set-lock=1 -C--kind="0.5.2" --sdist
   python -m build -C--set-lock=0 -C--kind="current" --sdist
   python -m build -C--set-lock=0 -C--kind='dog food tastes better than this' --sdist

   python -m --no-deps dist/[pck].whl

.. note:: imports work how?!

   In ``pyproject.toml`` section [build-system], :code:`backend-path=[".", "src"]`
   ``src/`` folder gets appended to the :code:`sys.path`

.. note:: BackendType subclasses

   Must be imported, after BackendType. This registers them

.. seealso::

   `in-tree backend <https://setuptools.pypa.io/en/latest/build_meta.html#dynamic-build-dependencies-and-other-build-meta-tweaks>`_

   `pep660 <https://peps.python.org/pep-0660/>`_

   https://github.com/pypa/setuptools_scm/blob/main/_own_version_helper.py

"""

from __future__ import annotations

import copy
import sys
from collections.abc import Mapping
from pathlib import Path

from setuptools import build_meta as _orig
from setuptools.build_meta import *  # noqa: F401 F403

from drain_swamp.monkey.config_settings import ConfigSettings
from drain_swamp.monkey.wrap_get_version import (
    SEM_VERSION_FALLBACK_SANE,
    write_to_file,
)


def get_requires_for_build_sdist(config_settings=None):
    """Before building package, runs plugins:

    - create/refresh .lnk files
    - set the version

    config_settings keys

    set-lock

    - 1 -- .lnk --> .lock

    - 0 -- .lnk --> .unlock

    kind

    tag, now, current, or a version str

    Usage

    .. code-block:: shell

       python -m build -C--set-lock=1 -C--kind="0.5.2" --sdist --verbose
       python -m build --config-setting="--set-lock=1" --config-setting="--kind=0.5.2" --verbose

    .. note:: setuptools_scm

       setuptools_scm hook occurs after our plugins. Would brutishly and
       clumsly to overwrite the version.

       Remove pyproject.toml [tool.setuptools_scm] section

       setuptools_scm generates WARNINGs w/ tracebacks; can be safely ignored

    .. note::

       :code:`python -m build` runs within a subprocess exit codes always 1

    """
    mod_path = "_req_links/backend.py"

    # Document usage and echo raw user input
    msg_info = (
        "-C--set-lock=1 to lock dependencies.\n"
        "-C--set-lock=0 to unlock dependencies.\n"
        "If not provided, looks at pyproject.toml lock state"
    )
    print(msg_info)

    msg_info = (
        """-C--kind empty gets exit code 8\n"""
        """-C--kind="current" gets current scm version. Writes\n"""
        """-C--kind="tag" gets version from version file\n"""
        """-C--kind="0.0.1" forces build a particular version\n"""
        "If not provided, same as tag"
    )
    print(msg_info)

    msg_info = f"INFO {mod_path} config_settings: {config_settings!r}"
    print(msg_info)

    if config_settings is None or not isinstance(config_settings, Mapping):
        # tox does not pass in config_settings, check DS_CONFIG_SETTINGS
        cs = ConfigSettings()
        d_config_settings_tox = cs.read()
        if d_config_settings_tox is not None:
            d_config_settings = copy.deepcopy(d_config_settings_tox)
        else:
            msg_exc = (
                "ERROR In build backend, expecting config_settings to be "
                f"Mapping got {config_settings!r} "
            )
            print(msg_exc)
            exit_code = 8
            sys.exit(exit_code)
    else:  # pragma: no cover
        d_config_settings = copy.deepcopy(config_settings)

    # Ensure an initial version file exists. The version within can be incorrect
    path_cwd = Path.cwd()
    path_f = path_cwd.joinpath("pyproject.toml")
    write_to_file(str(path_f), SEM_VERSION_FALLBACK_SANE, is_only_not_exists=True)

    from drain_swamp.monkey.wrap_infer_version import run_build_plugins

    run_build_plugins(d_config_settings)

    return _orig.get_requires_for_build_sdist(d_config_settings)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. py:data:: __all__
   :type: tuple[str]
   :value: ("BackendSetupTools",)

   Module's exports

.. py:data:: _logger
   :type: logging.Logger

   Module level logger

"""

from __future__ import annotations

import logging

from .backend_abc import BackendType
from .constants import g_app_name

__package__ = "drain_swamp"
__all__ = ("BackendSetupTools",)

_logger = logging.getLogger(f"{g_app_name}.backend_abc")


class BackendSetupTools(BackendType):
    """Support for setuptools

    Create an instance using the factory,
    :py:meth:`drain_swamp.backend_abc.BackendType.load_factory`

    Can read in the relevant ``pyproject.toml`` section for dependencies
    and optional-dependencies, another section details extras, and then
    rewrite the editable section

    The editable section should look like

    When dependency locked

    .. code-block:: text

       # @@@ editable
       dependencies = { file = ["requirements/prod.lock"] }
       optional-dependencies.pip = { file = ["requirements/pip.lock"] }
       optional-dependencies.pip_tools = { file = ["requirements/pip-tools.lock"] }
       optional-dependencies.dev = { file = ["requirements/dev.lock"] }
       optional-dependencies.manage = { file = ["requirements/manage.lock"] }
       optional-dependencies.docs = { file = ["docs/requirements.lock"] }
       # @@@ end

       version = {attr = "logging_strict._version.__version__"}

    When not dependency locked

    .. code-block:: text

       # @@@ editable
       dependencies = { file = ["requirements/prod.unlock"] }
       optional-dependencies.pip = { file = ["requirements/pip.unlock"] }
       optional-dependencies.pip_tools = { file = ["requirements/pip-tools.unlock"] }
       optional-dependencies.dev = { file = ["requirements/dev.unlock"] }
       optional-dependencies.manage = { file = ["requirements/manage.unlock"] }
       optional-dependencies.docs = { file = ["docs/requirements.unlock"] }
       # @@@ end

       version = {attr = "logging_strict._version.__version__"}

    Other ``pyproject.toml`` dynamic properties are not placed within
    the editable section

    In pyproject.toml,

    .. code-block:: text

       [tool.pipenv-unlock]
       extras = [
           "pip",
           "pip_tools",  # <-- underscore
           "dev",
           "manage",
           "docs",
       ]
       required = "requirements/prod.in"
       optionals = [
           "requirements/prod.in",
           "requirements/pip.in",
           "requirements/pip-tools.in",  # <-- hyphen
           "requirements/dev.in",
           "requirements/manage.in",
           "docs/requirements.in",
       ]

    - extras

      There maybe other .in/.lock/.unlock files, such as kit.in and
      tox.in, these are used by tox or CI/CD. These will also need
      .lock and .unlock files

    - required

      Even if the package has no required dependencies,
      .in/.lock/.unlocked files should still exist. By knowing which
      is the required dependencies related files, will know the others
      are optional, besides ``pins[.in|.lock|.unlock]``

    - optionals

      package optional-dependencies

      Can also be provided/overridden by passing in cli options

      Not allow cooresponding ``[.in|.lock|.unlock]`` will be
      optional-dependencies.

      File stems will have hyphens not underscore

      ``optional-dependencies.pip_tools = { file = ["requirements/pip-tools.unlock"] }``

      The file stem is hyphen but the dependency contains underscores

    .. py:attribute:: BACKEND_NAME
       :type: str
       :value: "setuptools"

       This backend driver is setuptools

    .. py:attribute:: PYPROJECT_TOML_SECTION_NAME
       :type: str
       :value: "tool.setuptools.dynamic"

       Dynamic section

       Not supported:

       - setup.cfg
       - setup.py
       - pyproject.toml, but not dynamic

       An editable section is created by start and end tokens. So not
       needed to know the name of the section. Left here for completeness

    :ivar d_pyproject_toml: pyproject.toml in one giant dict
    :vartype d_pyproject_toml: dict[str, typing.Any]
    :ivar path_config: Absolute path to pyproject.toml file
    :vartype path_config: pathlib.Path
    :ivar required:

       Default None. From cli, relative path to required dependencies .in file

    :vartype required: tuple[str, pathlib.Path] | None
    :ivar optionals:

       Default empty tuple. relative path of optional dependency ``.in`` files

    :vartype optionals: dict[str, pathlib.Path]
    :ivar parent_dir:

           folder path If provided and acceptable overrides attr, path_config

    :vartype parent_dir: pathlib.Path | None

    :raises:

       - :py:exc:`FileNotFoundError` -- package required dependency file is required

    """

    BACKEND_NAME = "setuptools"
    PYPROJECT_TOML_SECTION_NAME = "tool.setuptools.dynamic"

    def __init__(
        self,
        d_pyproject_toml,
        path_config,
        required=None,
        optionals={},
        parent_dir=None,
        additional_folders=(),
    ) -> None:
        super().__init__()
        self._path_config = path_config

        # may raise: FileNotFoundError
        self.load(
            d_pyproject_toml,
            required=required,
            optionals=optionals,
            parent_dir=parent_dir,
            additional_folders=additional_folders,
        )

    @property
    def backend(self) -> str:
        cls = type(self)
        return cls.BACKEND_NAME

    def compose_dependencies_line(self, suffix):
        """Compose required dependency line

        .. code-block:: text

           dependencies = { file = ["requirements/prod.lock"] }

        In this case,
        required is ``requirements/prod.in``
        suffix is ``.lock`` or ``lock`` or ``.unlock`` or ``unlock``

        :param suffix: File suffix. Either ``.lock`` or ``.unlock``
        :type suffix: str
        :returns: tool.setuptools.dynamic dependencies line
        :rtype: collections.abc.Iterator[str]
        """
        _logger.info(f"path_required: {self.required} {type(self.required)}")
        if self.required is None:
            yield from ()
        else:
            # tuple[str, Path]
            target, path_required_abs = self.required
            path_rel = path_required_abs.relative_to(self.parent_dir)
            path_dir = path_rel.parent

            str_suffix = self.fix_suffix(suffix)

            # file stem contains hyphens, not underscores
            stem = path_rel.stem
            stem = stem.replace("_", "-")

            path_dir_final = path_dir.joinpath(f"{stem}{str_suffix}")

            ret = f"""dependencies = {{ file = ["{path_dir_final}"] }}"""

            yield ret

        yield from ()

    def compose_optional_lines(self, suffix):
        """Compose the optional lines returning an Iterator.

        Optional lines are in this format

        .. code-block:: text

           optional-dependencies.pip_tools = { file = ["requirements/pip-tools.unlock"] }

        In this case,

        - suffix is ``.unlock``
        - extra is ``pip_tools``
        - optional is ``requirements/pip-tools.in``

        :param suffix: File suffix. Either ``.lock`` or ``.unlock``
        :type suffix: str
        :returns: tool.setuptools.dynamic dependencies and optional-dependencies lines
        :rtype: collections.abc.Iterator[str]
        """
        for target, path_abs in self.optionals.items():
            path_rel = path_abs.relative_to(self.parent_dir)
            str_suffix = self.fix_suffix(suffix)
            # file stem contains hyphens, not underscores
            path_rel_r = path_rel.parent
            stem_r = path_rel.stem
            stem_r = stem_r.replace("_", "-")
            path_full_r = path_rel_r.joinpath(f"{stem_r}{str_suffix}")

            target_l = target.replace("-", "_")
            ret = (
                f"""optional-dependencies.{target_l} = {{ file = ["""
                f""""{str(path_full_r)}"] }}"""
            )
            yield ret

        yield from ()

    def compose(self, suffix):
        """Create the new contents to be placed into the snippet.
        ``pyproject.toml`` required and optional-dependencies.

        :param suffix: File suffix. Either ``.lock`` or ``.unlock``
        :type suffix: str
        :returns: tool.setuptools.dynamic dependencies and optional-dependencies lines
        :rtype: str
        """
        super().compose()  # do validation checks
        ret = []
        ret.extend(list(self.compose_dependencies_line(suffix)))
        ret.extend(list(self.compose_optional_lines(suffix)))
        lines = "\n".join(ret)

        return lines


BackendType.register(BackendSetupTools)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Snip plugin for Sphinx ``conf.py``. Replaces snippet containing variables:

- copyright

- version

- release

- release_date

.. py:data:: __all__
   :type: tuple[str]
   :value: ("SnipSphinxConf", "entrypoint_name")

   Module exports

.. py:data:: entrypoint_name
   :type: str
   :value: "drain-swamp"

   Will use this entrypoint. In ``pyproject.toml``, ``[tool.drain-swamp]``

.. py:data:: _logger
   :type: logging.Logger

   Module level logger

"""

from __future__ import annotations

import datetime
import logging
import textwrap

from .constants import g_app_name
from .package_metadata import PackageMetadata
from .snip import (
    ReplaceResult,
    Snip,
)
from .version_semantic import (
    SemVersion,
    _path_or_cwd,
)

__all__ = ("SnipSphinxConf", "entrypoint_name")

entrypoint_name = "drain-swamp"

_logger = logging.getLogger(f"{g_app_name}.snippet_sphinx_conf")


class SnipSphinxConf:
    """Snip plugin for Sphinx conf.py snippet which sets:

    - copyright
    - version
    - release
    - release_date

    This is done for every tagged release

    Sphinx docs folder name is either ``doc`` or ``docs``. If it's something
    else, complain!

    .. py:attribute:: DOC_FOLDERS
       :type: tuple[str, str]
       :value: ("doc", "docs")

       Possible Sphinx doc folder basename

    :ivar path:

       Default None. absolute path to package base folder. If None, use
       current working folder

    :vartype path: pathlib.Path | None
    :raises:

       - :py:exc:`NotADirectoryError` -- From package base folder there
         is no doc/ or docs/ folder

       - :py:exc:`FileNotFoundError` -- In a doc folder, Sphinx config
         conf.py file not found

    """

    DOC_FOLDERS = ("doc", "docs")

    def __init__(self, path=None):
        super().__init__()

        # package base folder
        self.path_cwd = path

        # Sphinx conf python file
        self.path_abs_init()

        self._contents = None
        self._sv = None
        self._author_name_left = None

    @classmethod
    def now(cls):
        """Current datetime

        :returns: current datetime
        :rtype: datetime.datetime
        """
        return datetime.datetime.now()

    @classmethod
    def now_to_str(cls, strftime_str):
        """Format datetime now

        :param strftime_str: format str
        :type strftime_str: str
        :returns: datetime formatted with strftime format
        :rtype: str
        """
        dt_now = cls.now()
        try:
            ret = dt_now.strftime(strftime_str)
        except (ValueError, TypeError):
            raise

        return ret

    @property
    def path_abs(self):
        """Absolute path to ``doc/conf.py`` or ``docs/conf.py``

        :returns: package base folder
        :rtype: pathlib.Path
        """
        return self._path_abs

    def path_abs_init(self):
        """Using from package base path or cwd, find doc/ or docs/ folder absolute path

        :raises:

           - :py:exc:`NotADirectoryError` -- From package base folder
             there is no doc/ or docs/ folder

           - :py:exc:`FileNotFoundError` -- In a doc folder, Sphinx config
             conf.py file not found

        """
        cls = type(self)
        path_abs = None
        for path_child in self.path_cwd.iterdir():
            if path_child.is_dir() and path_child.name in cls.DOC_FOLDERS:
                path_abs = path_child
            else:  # pragma: no cover
                pass

        if path_abs is None:
            msg_exc = "There is no Sphinx docs or doc folder. Choose one and mkdir"
            raise NotADirectoryError(msg_exc)

        path_conf_py = path_abs.joinpath("conf.py")
        if not path_conf_py.exists() or (
            path_conf_py.exists() and not path_conf_py.is_file()
        ):
            msg_exc = "In a doc folder, Sphinx config conf.py file not found"
            raise FileNotFoundError(msg_exc)

        self._path_abs = path_conf_py

    @property
    def path_cwd(self):
        """Get package base folder

        :returns: package base folder
        :rtype: pathlib.Path
        """
        return self._path_cwd

    @path_cwd.setter
    def path_cwd(self, val):
        """Setter of current working directory

        :param val:

           None will get the current working directory. If not None,
           pass in a temp folder

        :type val: typing.Any | None
        """
        self._path_cwd = _path_or_cwd(val)

    @property
    def SV(self) -> SemVersion | None:
        """Getter for SemVersion instance

        :returns: After call to contents, SemVersion instance is available
        :rtype: SemVersion | None
        """
        return self._sv

    @property
    def author_name_left(self) -> str | None:
        """Getter for author name, but just the left portion.

        In Japanese, this is the full name. Otherwise would be the first name

        Used within regex to replace copyright notice

        :returns: Author name left. If contents yet to be called will be None
        :rtype: str | None
        """
        return self._author_name_left

    def contents(
        self,
        kind,
        package_name,
        copyright_start_year,
    ):
        """Create the snippet for Sphinx ``conf.py``. This will set:

        - copyright
        - version
        - release
        - release_date

        This is done for every tagged release

        Sphinx docs folder name is either ``doc`` or ``docs``. If it's something
        else, complain!

        :param kind:

           Explicit version str or "current"|"now" or "tag". ``tag`` for
           the latest tagged version. Current for the latest version
           which is probably the development version

        :type kind: str
        :param package_name:

           Can contain either hyphen or underscore. Both will work. From
           the package name can get the author name and email address. Only
           interested in the author name

        :type package_name: str
        :param copyright_start_year: This packages copyright start year
        :type copyright_start_year: str
        :raises:

           - :py:exc:`AssertionError` -- Expected Sphinx docs folder to be
             either ``doc/`` or ``docs/``

           - :py:exc:`ValueError` -- Explicit version str is invalid

        """
        cls = type(self)
        sv = SemVersion(path=self.path_cwd)
        path_project_base_dir = sv.path_cwd
        try:
            # Supply package name
            ver_full = sv.version_clean(kind)
        except (AssertionError, ValueError):
            self._contents = None
            raise

        # For use outside of the scope of Sphinx conf.py
        self._sv = sv

        """get authors name from package metadata then pyproject.toml
        Fallback --> "Anonymous"
        """
        author = PackageMetadata(package_name, path=path_project_base_dir)
        # For use outside of the scope of Sphinx conf.py
        self._author_name_left = author.left_name

        self._sv.parse_ver(ver_full)
        ver_xyz = self._sv.version_xyz()

        # \N{EN DASH} is a unicode char, elongated hyphen
        now_year = cls.now_to_str("%Y")
        # replaces f"""{self.now:%B %-d, %Y}"""
        now_full = cls.now_to_str("%B %-d, %Y")
        contents = textwrap.dedent(
            f"""\
            copyright = "{copyright_start_year}\N{EN DASH}{now_year}, {author.full_name}"
            # The short X.Y.Z version.
            version = "{ver_xyz}"
            # The full version, including alpha/beta/rc tags.
            release = "{ver_full}"
            # The date of release, in "monthname day, year" format.
            release_date = "{now_full}"
            """
        )
        self._contents = contents.rstrip()

    def replace(self, snippet_co=None):
        """In a Sphinx conf.py file, for a given snippet, replace contents

        :param snippet_co:

           Default None. id of snippet within Sphinx ``conf.py``. If
           None, that snippet better not have an id or there be other
           snippets in that file

        :type snippet_co: str | None
        :returns:

           - VALIDATE_FAIL -- Invalid snippet
           - NOMATCH -- no matching snippet with snippet_co
           - NO_CHANGE -- no replacement occurred
           - REPLACED -- replaced contents

        :rtype: drain_swamp.snip.ReplaceResult
        """
        # is_ok check avoids, TypeError
        contents = self._contents
        if contents is not None:
            snip = Snip(self.path_abs, is_quiet=True)
            ret = snip.replace(contents, id_=snippet_co)
        else:  # pragma: no cover
            ret = ReplaceResult.NO_CHANGE

        return ret



import logging
from collections.abc import (
    Callable,
    Iterable,
    Mapping,
    Sequence,
)
from types import ModuleType
from typing import Any

from pluggy import (
    HookImpl,
    PluginManager,
    Result,
)

logger: logging.Logger

__all__ = (
    "before",
    "after",
    "get_plugin_manager",
)

def before(
    hook_name: str,
    hook_impls: Sequence[HookImpl],
    kwargs: Mapping[str, Any],
) -> None: ...
def after(
    outcome: Result[Any],
    hook_name: str,
    hook_impls: Sequence[HookImpl],
    kwargs: Mapping[str, Any],
) -> None: ...
def lazy_dotted_path(dotted_path: str) -> ModuleType: ...
def lazy_package(
    mod_pkg: ModuleType,
) -> Sequence[ModuleType]: ...
def _filter_specs(member: Any) -> bool: ...
def iter_specs(
    mod: ModuleType,
    filter_: Callable[[Any], bool] = ...,
) -> Iterable[Any]: ...
def _create_hook_manager(namespace: str, specs_dotted_path: str) -> PluginManager: ...
def _register_hooks(
    hook_manager: PluginManager,
    hooks: Iterable[Any | tuple[Any, str | None]],
) -> None: ...
def _register_hooks_entry_points(
    hook_manager: PluginManager,
    entrypoint_plugins: str,
    disabled_plugins: Iterable[str],
) -> None: ...
def get_plugin_manager(
    namespace: str | None = ...,
    specs_dotted_path: str | None = ...,
    entrypoint_plugins: str | None = ...,
) -> PluginManager: ...



"""
https://github.com/pypa/setuptools_scm/blob/main/testing/wd_wrapper.py
"""

from __future__ import annotations

import itertools
from pathlib import Path
from typing import Any


class WorkDir:
    """a simple model for a"""

    commit_command: str
    signed_commit_command: str
    add_command: str

    def __repr__(self) -> str:
        return f"<WD {self.cwd}>"

    def __init__(self, cwd: Path) -> None:
        self.cwd = cwd
        self.__counter = itertools.count()

    def __call__(self, cmd: list[str] | str, **kw: object) -> str:
        if kw:
            assert isinstance(cmd, str), "formatting the command requires text input"
            cmd = cmd.format(**kw)
        from setuptools_scm._run_cmd import run

        return run(cmd, cwd=self.cwd).stdout

    def write(self, name: str, content: str | bytes) -> Path:
        path = self.cwd / name
        if isinstance(content, bytes):
            path.write_bytes(content)
        else:
            path.write_text(content, encoding="utf-8")
        return path

    def _reason(self, given_reason: str | None) -> str:
        if given_reason is None:
            return f"number-{next(self.__counter)}"
        else:
            return given_reason

    def add_and_commit(
        self, reason: str | None = None, signed: bool = False, **kwargs: object
    ) -> None:
        self(self.add_command)
        self.commit(reason=reason, signed=signed, **kwargs)

    def commit(self, reason: str | None = None, signed: bool = False) -> None:
        reason = self._reason(reason)
        self(
            self.commit_command if not signed else self.signed_commit_command,
            reason=reason,
        )

    def commit_testfile(self, reason: str | None = None, signed: bool = False) -> None:
        reason = self._reason(reason)
        self.write("test.txt", f"test {reason}")
        self(self.add_command)
        self.commit(reason=reason, signed=signed)

    def get_version(self, **kw: Any) -> str:
        __tracebackhide__ = True
        from setuptools_scm import get_version

        version = get_version(root=self.cwd, fallback_root=self.cwd, **kw)
        print(self.cwd.name, version, sep=": ")
        return version



import logging
from collections.abc import Callable
from typing import (
    Any,
    Final,
)

import setuptools_scm._types as _t
from setuptools_scm import (
    Configuration,
    ScmVersion,
)

log: logging.Logger

__all__ = (
    "scm_version",
    "write_to_file",
    "SEM_VERSION_FALLBACK_SANE",
)

SEM_VERSION_FALLBACK_SANE: Final[str]

try_parse: list[Callable[[_t.PathT, Configuration], ScmVersion | None]]

def _parse(root: str, config: Configuration) -> ScmVersion | None: ...
def scm_version(relative_to: str, sane_default: Any | None = ...) -> str: ...
def write_to_file(
    name: str,
    str_ver: str,
    write_to: str | None = None,
    dist_name: str | None = None,
) -> None: ...



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Get the scm version and writes it to version_file

.. py:data:: log
   :type: logging.Logger

   Module level logger

.. seealso::

   `kedro hooks <https://docs.kedro.org/en/latest/extend_kedro/plugins.html#hooks>`_
   `define impl <https://docs.kedro.org/en/latest/hooks/introduction.html#define-the-hook-implementation>`_


Alternative implementation -- Same as setuptools_scm version_infer
Raises LookupError if missing pyproject.toml or missing sections tool.drain-swamp and tool.pipenv-unlock

.. code-block:: text

   from setuptools_scm._integration.setuptools import infer_version as _infer_version
   msg_warn = (
       f"{mod_path} drain-swamp plugins not being run! This should be an error "
       "and stop the build"
   )
   log.warning(msg_warn)
   # setuptools_scm._integration.setuptools._assign_version
   # setuptools_scm._config.Configuration.from_file
   with (
       patch(
           "setuptools_scm._integration.setuptools._config._read_pyproject",
           new_callable=MagicMock(wraps=ReadPyprojectStrict),
       ),
       patch(
           "setuptools_scm._config._read_pyproject",
           new_callable=MagicMock(wraps=ReadPyprojectStrict),
       ),
   ):
       # Limited to only setuptools_scm._config.Configuation fields
       _infer_version(dist)


"""

from __future__ import annotations

import logging
import sys
import warnings
from importlib.metadata import (
    PackageNotFoundError,
    metadata,
)
from pathlib import Path
from typing import Any

from drain_swamp._run_cmd import run_cmd
from drain_swamp.monkey.config_settings import ConfigSettings
from drain_swamp.monkey.hooks import markers
from drain_swamp.monkey.hooks.constants import HOOK_NAMESPACE

log = logging.getLogger(__name__)


def _is_package_installed(app_name):
    """Check if package is installed

    :param app_name: Accepts either package name or app name. No need for check
    :type app_name: str
    :returns: True package is installed otherwise False
    :rtype: bool
    """
    # metadata is hardened; accepts either package name or app name
    try:
        metadata(app_name)
    except PackageNotFoundError:
        ret = False
    else:
        ret = True

    return ret


def _kind(config_settings, fallback="tag"):
    """Sanitize kind user input

    :param config_settings: config dict
    :type config_settings: dict[str, typing.Any] | None
    :param fallback: Default "tag". kind default
    :type fallback: str | None
    :returns: kind can be: current, now, tag, or a semantic version str
    :rtype: str
    """
    if fallback is None or not isinstance(fallback, str):
        str_fallback = "tag"
    else:  # pragma: no cover
        str_fallback = fallback

    is_not_dict = config_settings is None or not isinstance(config_settings, dict)
    if is_not_dict:
        ret = str_fallback
    else:
        # config_settings is coming from the cli, values will be str
        try:
            if "--kind" in config_settings.keys():
                ret = config_settings["--kind"]
            elif "kind" in config_settings.keys():
                ret = config_settings["kind"]
            else:
                # Do not use scm version. Version from tool.pipenv-unlock.version_file
                ret = str_fallback

            # whitespace only is invalid --> AssertionError
            assert isinstance(ret, str) and len(ret.strip()) != 0
        except AssertionError:
            msg_warn_invalid_kind = (
                "-C--kind expected to be: tag, current, now, or a "
                f"version str. Got {ret!r}"
            )
            warnings.warn(msg_warn_invalid_kind)
            ret = str_fallback

    return ret


@markers.hook_impl(specname=f"{HOOK_NAMESPACE}_on_version_infer")
def on_version_infer(config_settings: dict[str, Any]) -> str | None:
    """Update version_file depending on kind: current, now, tag,
    semantic version str

    pluggy will inspect this method's signature. So keep it!
    """
    mod_path = "backend plugin ds_scm_version"

    # Use absolute path. Relative path would silently fail
    p_bin_dir = Path(sys.executable).parent
    scm_version_path = str(p_bin_dir.joinpath("scm-version"))

    is_installed = _is_package_installed("drain_swamp")
    ret = None
    with warnings.catch_warnings(record=True) as w:
        kind = _kind(config_settings)
        if len(w) != 0:
            ret = str(w[0].message)

    log.info(f"{mod_path} kind: {kind!r}")
    config_settings_path = ConfigSettings.get_abs_path()
    log.info(f"{mod_path} config_settings_path: {config_settings_path}")

    if ret is not None:
        return ret
    else:
        if kind in ("current", "now"):
            # Get current version from scm and write to file
            if not is_installed:  # pragma: no cover
                # pep366
                cmd = [
                    sys.executable,
                    "src/drain_swamp/cli_scm_version.py",
                ]
            else:
                # ../cli_scm_version.py entrypoint
                cmd = [
                    scm_version_path,
                ]
            cmd.extend(["get", "--is-write"])
            t_out = run_cmd(cmd, cwd=None)
            out, err, code, subprocess_msg = t_out
            if subprocess_msg is not None:
                ret = None
                warnings.warn(subprocess_msg)
            else:
                # A plugins output should be None if no error occurred
                if code == 0:
                    ret = None
                else:
                    # error occurred
                    if err is None:  # pragma: no cover
                        # on stderr, entrypoint did not provide an explanation
                        # make an adhoc msg
                        ret = f"Exit code: {code} cmd {cmd}"
                    else:
                        # on stderr, entrypoint provided an explanation
                        ret = err
        elif kind == "tag":
            """:code:`python -m build` will use version from
            version_file. Do not change version_file"""
            ret = None
        else:
            # version str
            if not is_installed:  # pragma: no cover
                # pep366
                cmd = [
                    sys.executable,
                    "src/drain_swamp/cli_scm_version.py",
                ]
            else:
                # ../cli_scm_version.py entrypoint
                cmd = [
                    scm_version_path,
                ]
            cmd.extend(["write", kind])
            t_out = run_cmd(cmd, cwd=None)
            out, err, code, subprocess_msg = t_out
            if subprocess_msg is not None:
                ret = None
                warnings.warn(subprocess_msg)
            else:
                # A plugins output should be None if no error occurred
                if code == 0:
                    ret = None
                else:
                    # error occurred
                    if err is None:  # pragma: no cover
                        # on stderr, entrypoint did not provide an explanation
                        # make an adhoc msg
                        ret = f"Exit code: {code} cmd {cmd}"
                    else:
                        # on stderr, entrypoint provided an explanation
                        ret = err

            # for debugging purposes
            # ret = repr(t_out)
            pass

        return ret



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Package wide exceptions

.. py:data:: __all__
   :type: tuple[str, str, str, str]
   :value: ("PyProjectTOMLParseError", "BackendNotSupportedError", \
   "PyProjectTOMLReadError", "MissingRequirementsFoldersFiles")

   Module exports

"""

__package__ = "drain_swamp"
__all__ = (
    "PyProjectTOMLParseError",
    "BackendNotSupportedError",
    "PyProjectTOMLReadError",
    "MissingRequirementsFoldersFiles",
)


class PyProjectTOMLReadError(OSError):
    """There is an issue with the ``pyproject.toml``. Not a file or
    insufficient permissions

    Although an :py:exc:`OSError`, it's generic. The cause of the error
    is not passed forward (aka allowed to be lost)

    :ivar msg: The error message
    :vartype msg: str
    """

    def __init__(self, msg: str) -> None:
        errno = None
        super().__init__(errno, msg)


class PyProjectTOMLParseError(ValueError):
    """No point in continuing if the ``pyproject.toml`` can't be parsed and loaded

    :ivar msg: The error message
    :vartype msg: str
    """

    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class BackendNotSupportedError(ValueError):
    """Do not have support yet for given Python package backend

    :ivar msg: The error message
    :vartype msg: str
    """

    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class MissingRequirementsFoldersFiles(AssertionError):
    """Neglected to create/prepare requirements folders and ``.in`` files

    Unabated would produce an empty string snippet. Instead provide
    user feedback

    :ivar msg: The error message
    :vartype msg: str
    """

    def __init__(self, msg: str) -> None:
        super().__init__(msg)
